# %%
import pandas as pd
import numpy as np
import time
import datetime
from datetime import datetime as dt
from datetime import timezone
from spacepy import coordinates as coord
from spacepy.time import Ticktock
from astropy.constants import R_earth
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.offline import iplot
import kamodo
from flythrough.utils import ConvertCoord
from kamodo_ccmc.tools.shoreline import shoreline

def SatPlot4D(var,time,c1,c2,c3,vard,varu,inCoordName,inCoordType,plotCoord,groupby,model,
              displayplot=True,returnfig=False,type='3D',body='black',zoom=False,divfile='',
              htmlfile='',plotCoordType1D='car',
              vUnit='',vxName='',vx='',vyName='',vy='',vzName='',vz='',vScale=1.,vSkip=0):
    """4D plotting for satellite trajectories using plotly by Darren De Zeeuw
    
    __Required variables__
    
    var: string of variable name
    time: time formatted as a timestamp in UTC
    c1: latitude  or X
    c2: longitude or Y
    c3: altitude  or Z
    vard: data of variable var, same size array as positions
    varu: string of variable var units
    inCoordName: string for incoming coordinate system.
        GDZ, GEO, GSM, GSE, SM, GEI, MAG, RLL
    inCoordType: string for incoming coordinate type.  car, sph
    plotCoord: string for coordinate system used in 3D plot. 
        Assumes car for 3D and Polar, sph otherwise.
    groupby: grouping of data for animation, values include
        all, day, hour, minute, N, orbitE, orbitM
    model: string of name of model the data was extracted from
    
    __Optional variables__
    
    displayplot: logical to show/hide displayed plot
    returnfig: logical to return figure object for further modification,
        will override displayplot if True
    type: string for choice of plot type, values:
        3D, 3Dv, 1D, 2D, 2DLT, 2DPN, 2DPS
    body: string choice of 3D inner body, values: black, earth (GEO), none
    zoom: logical to show zoomed in view for polar plots
    divfile: string with filename to save a html div file of the plot
    htmlfile: string with filename to save a full html file of the plot
    plotCoordType1D: displayed coordinate type for 1D plots.  car, sph
    vUnit: units of quiver vector quantity, ie 'nT'
    vxName: string name of X component of quiver vector, ie 'B_x'
    vx: array of values of X component of quiver vector
    vyName: string name of Y component of quiver vector, ie 'B_y'
    vy: array of values of Y component of quiver vector
    vzName: string name of Z component of quiver vector, ie 'B_z'
    vz: array of values of Z component of quiver vector
    vScale: length scale factor to quiver vectors
    vSkip: skip this number of points between displaying quiver vector
    """

    REkm = (R_earth.value/1000.)

    import kamodo_ccmc.flythrough.model_wrapper as MW
    iC_names=MW.coord_names(inCoordName,inCoordType)
    iC_units=MW.coord_units(inCoordName,inCoordType)
    if inCoordType=='sph':
        for key in iC_names.keys(): iC_names[key]=iC_names[key][:3]
    oC_names=MW.coord_names(plotCoord,plotCoordType1D)
    if plotCoordType1D=='sph':
        for key in oC_names.keys(): oC_names[key]=oC_names[key][:3]
    else:
        for key in oC_names.keys(): oC_names[key]=oC_names[key][:1]

    isQuiver = False
    if type == "3Dv":  isQuiver = True

    if type == "3D" or type == "3Dv" or type == "2DPN" or type == "2DPS":
        #Convert incoming coordinates into plot coordinages (cartesian)
        xx,yy,zz,units = ConvertCoord(time,c1,c2,c3,inCoordName,inCoordType,plotCoord,'car')

        # Create dictionary block to pass to plotting with selected options
        plot_dict=dict(
            title = 'Satellite extraction from model: '+model+"<br>"+plotCoord+" coordinates",
            sats = ["Sat1"],
            Sat1 = dict(
                display_name = "",
                time = dict(format='timestamp', data=time),
                vars = dict(
                    x = dict(units=units[0], data=xx, coord=plotCoord),
                    y = dict(units=units[1], data=yy, coord=plotCoord),
                    z = dict(units=units[2], data=zz, coord=plotCoord)
                ),
                position_variables = ["x", "y", "z"],
            ),
            options = dict(
                position_units = "R_E",
                var = var,
                hover_vars = [],
                quiver = isQuiver,
                quiver_scale = vScale,
                quiver_skip = vSkip,
                groupby = groupby,
                body = body,
                colorscale = "Viridis",
                REkm = REkm,
                coord = plotCoord,
            ),
        )
        # Fixed position variables already included, now add passed in variable to dictionary
        plot_dict['Sat1']['vars'][var]=dict(units=varu, data=vard, coord=plotCoord)
        plot_dict['Sat1']['vars'][iC_names['c1']]=dict(units=iC_units['c1'], data=c1, coord=inCoordName)
        plot_dict['Sat1']['vars'][iC_names['c2']]=dict(units=iC_units['c2'], data=c2, coord=inCoordName)
        plot_dict['Sat1']['vars'][iC_names['c3']]=dict(units=iC_units['c3'], data=c3, coord=inCoordName)
        if plotCoord != inCoordName:
            plot_dict['options']['hover_vars']=[iC_names['c1'],iC_names['c2'],iC_names['c3']]
        if isQuiver:
            plot_dict['Sat1']['vars'][vxName]=dict(units=vUnit, data=vx, coord=inCoordName)
            plot_dict['Sat1']['vars'][vyName]=dict(units=vUnit, data=vy, coord=inCoordName)
            plot_dict['Sat1']['vars'][vzName]=dict(units=vUnit, data=vz, coord=inCoordName)
            plot_dict['Sat1']['vector_variables']=[vxName,vyName,vzName]

        # Execute creation and display of figure
        if type == "3D" or type == "3Dv":
            fig=custom3Dsat(plot_dict,vbose=0)
        elif type == "2DPN":
            fig=custom2Dpolar(plot_dict,'N',zoom=zoom,vbose=0)
        elif type == "2DPS":
            fig=custom2Dpolar(plot_dict,'S',zoom=zoom,vbose=0)
        if divfile != '':
            print('-saving html div file: ',divfile)
            fig.write_html(divfile,full_html=False)
        if htmlfile != '':
            print('-saving full html file: ',htmlfile)
            fig.write_html(htmlfile,full_html=True)
        if returnfig:
            return fig
        else:
            if displayplot:
                iplot(fig)

    if type == "1D" or type == "2D" or type == "2DLT":
        newCoord = False
        if type is "1D" and ( inCoordName != plotCoord or inCoordType != plotCoordType1D):
            #Convert incoming coordinates into plot coordinages (cartesian)
            xx,yy,zz,units = ConvertCoord(time,c1,c2,c3,inCoordName,inCoordType,plotCoord,plotCoordType1D)
            newCoord = True

        # Create dictionary block to pass to plotting with selected options
        plot_dict=dict(
            title = 'Satellite extraction from model: '+model+"<br>"+plotCoord+" coordinates",
            sats = ["Sat1"],
            Sat1 = dict(
                display_name = "",
                time = dict(format='timestamp', data=time),
                vars = dict(),
                position_variables = [],
            ),
            options = dict(
                position_units = "",
                var = var,
                hover_vars = [],
                groupby = groupby,
                coord = plotCoord,
            ),
        )
        # Fixed position variables already included, now add passed in variable to dictionary
        plot_dict['Sat1']['vars'][var]=dict(units=varu, data=vard, coord=plotCoord)
        if newCoord:
            plot_dict['Sat1']['vars'][oC_names['c1']]=dict(units=units[0], data=xx, coord=plotCoord)
            plot_dict['Sat1']['vars'][oC_names['c2']]=dict(units=units[1], data=yy, coord=plotCoord)
            plot_dict['Sat1']['vars'][oC_names['c3']]=dict(units=units[2], data=zz, coord=plotCoord)
            plot_dict['Sat1']['position_variables']=[oC_names['c1'],oC_names['c2'],oC_names['c3']]
            plot_dict['options']['hover_vars']=[oC_names['c1'],oC_names['c2'],oC_names['c3']]
        else:
            plot_dict['Sat1']['vars'][iC_names['c1']]=dict(units=iC_units['c1'], data=c1, coord=inCoordName)
            plot_dict['Sat1']['vars'][iC_names['c2']]=dict(units=iC_units['c2'], data=c2, coord=inCoordName)
            plot_dict['Sat1']['vars'][iC_names['c3']]=dict(units=iC_units['c3'], data=c3, coord=inCoordName)
            plot_dict['Sat1']['position_variables']=[iC_names['c1'],iC_names['c2'],iC_names['c3']]
            plot_dict['options']['hover_vars']=[iC_names['c1'],iC_names['c2'],iC_names['c3']]

        # Execute creation and display of figure
        if type == "1D":
            fig=custom1Dsat(plot_dict,vbose=0)
        elif type == "2D":
            fig=custom2Dsat(plot_dict,vbose=0)
        elif type == "2DLT":
            fig=custom2Dsat(plot_dict,useLT=True,vbose=0)

        if divfile != '':
            print('-saving html div file: ',divfile)
            fig.write_html(divfile,full_html=False)
        if htmlfile != '':
            print('-saving full html file: ',htmlfile)
            fig.write_html(htmlfile,full_html=True)
        if returnfig:
            return fig
        else:
            if displayplot:
                iplot(fig)


# ===============================================================================================
# ===============================================================================================
def custom3Dsat(datad, vbose=1):
    '''
    This function creates a custom 3D satellite plot, returning a plotly figure object.
    
    Parameters
    ----------
    datad: This is a data dictionary with the data used to create the plot
    vbose: An optional verbosity value, 0 will only print out warnings and errors. Default is 1.
    
    Returns
    -------
    fig: A plotly figure object that can then be visualized.
    
    Other
    -----
    The python code block below will setup and display a working demo.

import numpy as np
import datetime
from datetime import timezone
from plotly.offline import iplot

# Build a datetime array to use in the dictionary
base = datetime.datetime(2000, 1, 1).replace(tzinfo=timezone.utc)
arr = np.array([base + datetime.timedelta(minutes=30*i) for i in range(8)])

sample=dict(
    title = 'Plot Title Here',  # Displayed title for plot, can use <br> for new lines
    sats = ["Sat1"],            # Array of satellites to include in plot
    Sat1 = dict(
        display_name = "Fake Satellite",
        time = dict(format='datetime', data=arr),  # possible formats: datetime, timestamp (assumes UTC)
        vars = dict(
            x = dict(units='R_E', data=np.array(np.arange(1.,9.))),
            y = dict(units='R_E', data=np.array(np.arange(1.,9.))),
            z = dict(units='R_E', data=np.array(np.arange(1.,9.))),
            p = dict(units='nP', data=np.array(np.arange(11.,19.))),
            U_x = dict(units='km/s', data=np.array(-1.*np.arange(11.,19.))),
            U_y = dict(units='km/s', data=np.array(np.arange(21.,29.))),
            U_z = dict(units='km/s', data=np.array(np.arange(31.,39.))),
        ),
        position_variables = ["x", "y", "z"],     # three variables to use for position
        vector_variables = ["U_x", "U_y", "U_z"], # three variables to use for quiver if quiver is True
    ),
    options = dict(
        position_units = "R_E", # possible values: R_E, km, ""
        var = "p",              # variable to use for colorscale
        hover_vars = ["U_x"],   # other information for hoverinfo display
        quiver = True,          # logical value to display or hide quivers
        quiver_scale = 0.1,     # length scale of quivers
        quiver_skip = 0,        # points to skip between displaying quivers
        groupby = "orbitM",     # possible values: all, day, hour, minute, N (integer, show N values at a time)
                                #   orbitE (break at equator crossing S->N), orbitM (break at prime meridian crossing)
        body = "black",         # possible values: black, earth, and any other value is no body
        colorscale = "Viridis", # named colorscale
        REkm = 6.3781E3,        # Earth radius in km
    ),
)

fig=custom3Dsat(sample)
iplot(fig)
    '''

    # ===============================================================================================  
    # Start timer
    tic = time.perf_counter()

    # Start with error checking ...
    if 'title' not in datad:
        print("Warning, no title given for plot.")
        txttop = "No Title"
    else:
        txttop = datad['title']
    if 'var' not in datad['options']:
        print("ERROR, no variable selected to plot, returning.")
        return None
    var=datad['options']['var']
    ### NOTE: the checks on variable "time" are changed to "zzztime" throughout as they don't work as intended.
    if var == "zzztime":
        varu=""
    else:
        varu=datad[datad['sats'][0]]['vars'][var]['units']
    if 'REkm' in datad['options']:
        REkm = datad['options']['REkm']
    else:
        REkm=6.3781E3
    scale=datad['options']['position_units']
    if 'groupby' in datad['options']:
        groupby = datad['options']['groupby']
    else:
        groupby = "all"
    if 'quiver' in datad['options']:
        quiver=datad['options']['quiver']
    else:
        quiver=False
    if quiver:
        quiverscale=datad['options']['quiver_scale']
        if scale == "km":
            quiverscale=quiverscale*REkm
        quiverscale2=quiverscale
        quiverskip=int(datad['options']['quiver_skip'])
    if 'body' in datad['options']:
        body=datad['options']['body']
    else:
        body = "none"
    if 'colorscale' in datad['options']:
        colorscale = datad['options']['colorscale']
    else:
        colorscale = "Viridis"
    if 'coord' in datad['options']:
        coord = datad['options']['coord']
    else:
        coord = ""

    # set initial values used later, including loop over all sats
    xmin=0.
    xmax=0.
    ymin=0.
    ymax=0.
    zmin=0.
    zmax=0.
    cmin= 1.e99
    cmax=-1.e99
    localts=dict()
    localtimestring=dict()
    agroup=dict()
    ugroup=()
    for sat in datad['sats']:
        sPts=len(datad[sat]['vars'][datad[sat]['position_variables'][0]]['data'])
        # Set localtimestring values
        notime=False
        if 'time' in datad[sat]:
            if datad[sat]['time']['format'] == "datetime":
                localts[sat]=np.array([d.timestamp() for d in datad[sat]['time']['data']])
            elif datad[sat]['time']['format'] == "timestamp":
                localts[sat]=datad[sat]['time']['data'].copy()
            else:
                print("ERROR, Unknown time format.")
                return None
            localtimestring[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M:%S") for d in localts[sat]])
        else:
            notime=True
            if var == "zzztime":
                print("ERROR, no time given and plot var selected is time")
                return None
            localtimestring[sat]=np.array(["point "+str(i+1) for i in range(sPts)])

        # Find global contour min/max
        if var == "zzztime":
            c=localts[sat]
        else:
            c=datad[sat]['vars'][var]['data']
        cmin=min(cmin,min(c))
        cmax=max(cmax,max(c))

        # Create array of possible 'groupby' value
        if groupby == "day":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d") for d in localts[sat]])
        elif groupby == "hour":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H") for d in localts[sat]])
        elif groupby == "minute":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M") for d in localts[sat]])
        elif groupby == "orbitM":
            # Satellite path crosses prime meridian
            x=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            y=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            bgroup = ['orbit'] * len(x)
            j=1
            for i in range(sPts):
                if i != 0:
                    if x[i] > 0. and (y[i]*y[i-1]) < 0.:
                        j+=1
                bgroup[i] = "orbit "+str(j)
            agroup[sat]=np.array(bgroup)
        elif groupby == "orbitE":
            # Satellite path crosses equator going North
            z=datad[sat]['vars'][datad[sat]['position_variables'][2]]['data']
            bgroup = ['orbit'] * len(z)
            j=1
            for i in range(sPts):
                if i != 0:
                    if (z[i]>0. and z[i-1]<0.):
                        j+=1
                bgroup[i] = "orbit "+str(j)
            agroup[sat]=np.array(bgroup)
        elif groupby.isdigit():
            gb=int(groupby)
            agroup[sat] = np.array(["points "+str(int(i/gb)*gb+1)+" - "+str(int(i/gb)*gb+gb) for i in range(sPts)])
        else:
            agroup[sat] = np.array(["all" for i in range(sPts)])

        # Use pandas unique function rather than numpy. Its faster and does not sort the results.
        ugroup=pd.unique(np.append(ugroup, pd.unique(agroup[sat])))

    ngroup = len(ugroup)

    # Build DUMMY data block to insert as needed.
    data_dict_dummy = {
        "type": "scatter3d",
        "name": "dummy", "x": [0.], "y": [0.], "z": [0.],
        "mode": "lines", "line": {"width": 1},
        "hoverinfo": "none",
    }

    # 3D===============================================================================================  AAA
    # make figure dictionary pieces
    fig_dict = {"data": [], "layout": {}, "frames": []}
    fig_data_saved = {"data": []}
    sliders_dict = {
        "active": 0,
        "yanchor": "top", "xanchor": "left",
        "currentvalue": {
            "prefix": "Currently showing: ",
            "visible": True,
            "xanchor": "left"
        },
        "transition": {"duration": 0},
        "pad": {"b": 10, "t": 10},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": []
    }

    # Actual plot creation loop
    for date in ugroup:
        frame = {"data": [], "name": date}
        for sat in datad['sats']:
            x=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            y=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            z=datad[sat]['vars'][datad[sat]['position_variables'][2]]['data']
            sc=1.
            if scale == "km" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "R_E":
                sc=REkm
            elif scale == "R_E" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "km":
                sc=1./REkm
            if var == "zzztime":
                c=localts[sat]
                varline=""
            else:
                c=datad[sat]['vars'][var]['data']
                varline=var+": %{marker.color:.4g} "+varu+"<br>"
            if quiver:
                qxvar=datad[sat]['vector_variables'][0]
                qyvar=datad[sat]['vector_variables'][1]
                qzvar=datad[sat]['vector_variables'][2]
                qx=datad[sat]['vars'][qxvar]['data']
                qy=datad[sat]['vars'][qyvar]['data']
                qz=datad[sat]['vars'][qzvar]['data']
                qunit=datad[sat]['vars'][qxvar]['units']
                qcoord=datad[sat]['vars'][qxvar]['coord']
                # Set to unit scale
                qmax = np.max(np.sqrt(qx*qx + qy*qy + qz*qz))
                quiverscale2 = quiverscale/qmax

            # Update position min/max values
            if date == ugroup[0]:
                xmin=min(xmin,min(x*sc))
                xmax=max(xmax,max(x*sc))
                ymin=min(ymin,min(y*sc))
                ymax=max(ymax,max(y*sc))
                zmin=min(zmin,min(z*sc))
                zmax=max(zmax,max(z*sc))

            # Compute mask to restrict all data in trace
            mask = date == agroup[sat]
            
            # Create hover information, including extras passed in. Quiver shows additional variables.
            Nhv = len(datad['options']['hover_vars'])
            cd=[]
            cd.append(localtimestring[sat][mask])
            qline=""
            Ndv=1
            if quiver:
                cd.append(qx[mask])
                cd.append(qy[mask])
                cd.append(qz[mask])
                qline+=qxvar+": %{customdata[1]:.2f} "+qunit+" "+qcoord+"<br>"
                qline+=qyvar+": %{customdata[2]:.2f} "+qunit+" "+qcoord+"<br>"
                qline+=qzvar+": %{customdata[3]:.2f} "+qunit+" "+qcoord+"<br>"
                Ndv+=3
            for i in range(Nhv):
                cd.append(datad[sat]['vars'][datad['options']['hover_vars'][i]]['data'][mask])
                qline+=datad['options']['hover_vars'][i]+": %{customdata["+str(Ndv)+"]:.2f} "+\
                    datad[sat]['vars'][datad['options']['hover_vars'][i]]['units']+" "+\
                    datad[sat]['vars'][datad['options']['hover_vars'][i]]['coord']+"<br>"
                Ndv+=1
            cd=np.asarray(cd).T
            dateline="%{customdata[0]}<br>"
            
            # Build data block with mask
            data_dict = {
                "type": "scatter3d",
                "name": date,
                "x": list(x[mask]*sc), "y": list(y[mask]*sc), "z": list(z[mask]*sc),
                "mode": "markers+lines",
                "marker": {
                    "size": 4, "cmin": cmin, "cmax": cmax, "color": list(c[mask]),
                    "showscale": True, "colorscale": colorscale,
                    "colorbar": { "title": "<b>"+var+"</b><br>["+varu+"]", "tickformat": ".3g" }
                },
                "line": {"width": 3, "color": "rgba(22,22,22,0.2)"},
                "customdata": cd,
                "hovertemplate": "<b>"+datad[sat]['display_name']+"</b>"+
                    "<br>X: %{x:.4f} "+scale+"<br>Y: %{y:.4f} "+scale+"<br>Z: %{z:.4f} "+scale+"<br>"+
                    qline+varline+dateline+"<extra></extra>",
            }

            # If time is colorbar variable, hide labels by selecting ticks out of range
            if var == "zzztime":
                data_dict.marker.colorbar['tickvals']=(0,1)

            # Put each part of sequence in frame data block
            frame["data"].append(data_dict)

            # First in sequence, put dummy in main data block
            if date == ugroup[0]:
                fig_dict["data"].append(data_dict_dummy)
            
            # Start quiver
            if quiver:
                # Compute values to put in quiver trace
                # Make array max possible size (triple len(x)), fill, and trim as needed
                xx=np.concatenate([x,x,x])
                yy=np.concatenate([y,y,y])
                zz=np.concatenate([z,z,z])
                qxx=qx
                qyy=qy
                qzz=qz
                # Build new position array, element by element
                j=0
                for i in range(len(mask)):
                    if mask[i]:
                        if i%(quiverskip+1) == 0:
                            xx[j]=x[i]*sc
                            yy[j]=y[i]*sc
                            zz[j]=z[i]*sc
                            xx[j+1]=x[i]*sc+quiverscale2*qx[i]
                            yy[j+1]=y[i]*sc+quiverscale2*qy[i]
                            zz[j+1]=z[i]*sc+quiverscale2*qz[i]
                            xx[j+2]=None
                            yy[j+2]=None
                            zz[j+2]=None
                            j+=3
                xx=np.array(xx[0:j], dtype=np.float64)
                yy=np.array(yy[0:j], dtype=np.float64)
                zz=np.array(zz[0:j], dtype=np.float64)

                # Update position min/max values
                xmin=min(xmin,min(xx))
                xmax=max(xmax,max(xx))
                ymin=min(ymin,min(yy))
                ymax=max(ymax,max(yy))
                zmin=min(zmin,min(zz))
                zmax=max(zmax,max(zz))
                
                # Build data block
                data_dict = {
                    "type": "scatter3d",
                    "name": "positions", "x": list(xx), "y": list(yy), "z": list(zz),
                    "mode": "lines", "line": {"width": 3, "color": "rgba(55,22,22,0.4)"},
                    "hoverinfo": "none",
                }
                # Put each part of sequence in frame data block
                frame["data"].append(data_dict)
                
                # First in sequence, put in main data block
                if date == ugroup[0]:
                    fig_dict["data"].append(data_dict_dummy)
                
        fig_dict["frames"].append(frame)
        slider_step = {"args": [
            [date],
            {"frame": {"duration": 300, "redraw": True},
             "mode": "immediate",
             "transition": {"duration": 0}}
        ],
            "label": date,
            "method": "animate"}
        sliders_dict["steps"].append(slider_step)
        
    # Assemble frame and slider pieces
    fig_dict["layout"]["sliders"] = [sliders_dict]

    # 3D===============================================================================================  BBB
    if ngroup > 1:
        for sat in datad['sats']:
            # Add trace if more than one group.
            # This shows the whole trajectory when a subsection of data is showing.
            x=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            y=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            z=datad[sat]['vars'][datad[sat]['position_variables'][2]]['data']
            sc=1.
            if scale == "km" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "R_E":
                sc=REkm
            elif scale == "R_E" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "km":
                sc=1./REkm
            
            # Build data block
            data_dict = {
                "type": "scatter3d",
                "name": "positions", "x": list(x*sc), "y": list(y*sc), "z": list(z*sc),
                "mode": "lines", "line": {"width": 3, "color": "rgba(22,22,22,0.2)"},
                "hoverinfo": "none",
            }

            # Put into main data block
            fig_dict["data"].append(data_dict)

    # 3D===============================================================================================  CCC
    ticE = time.perf_counter()
    # Load points and add 1 RE sphere, padded to cover all data positions
    if body == "black":
        dataXYZ = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereXYZ.csv')
        dataIJK = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereIJK.csv')
        if scale == "km":
            dataXYZ *= REkm
        # Build data block
        data_dict = {
            "type": "mesh3d",
            "name": '1 R_E sphere',
            "x": list(np.append(dataXYZ['x'],(xmin,xmax))),
            "y": list(np.append(dataXYZ['y'],(ymin,ymax))),
            "z": list(np.append(dataXYZ['z'],(zmin,zmax))),
            "i": list(dataIJK['i']),
            "j": list(dataIJK['j']),
            "k": list(dataIJK['k']),
            "facecolor": list(dataIJK['c']),
            "flatshading": True,
            "hovertemplate": "Earth<extra></extra>",
        }
        # Put in main data block
        fig_dict["data"].append(data_dict)
    elif body == "earth" and coord == "GEO":
        dataXYZ = pd.read_csv('https://ccmc.gsfc.nasa.gov/ungrouped/GM_IM/EarthXYZ.csv')
        dataIJK = pd.read_csv('https://ccmc.gsfc.nasa.gov/ungrouped/GM_IM/EarthIJKRGB.csv')
        if scale == "km":
            dataXYZ *= REkm
        color=np.array(["rgb("+str(dataIJK['r'][i])+","+str(dataIJK['g'][i])+","+str(dataIJK['b'][i])+")" \
                        for i in range(len(dataIJK['r']))])
        # Need to reverse x,y from file to be in proper GEO coords (180 degree rotation)
        xe=-dataXYZ['x']
        ye=-dataXYZ['y']
        ze= dataXYZ['z']
        # Build data block
        data_dict = {
            "type": "mesh3d",
            "name": '1 R_E sphere',
            "x": np.append(xe,(xmin,xmax)),
            "y": np.append(ye,(ymin,ymax)),
            "z": np.append(ze,(zmin,zmax)),
            "i": dataIJK['i'],
            "j": dataIJK['j'],
            "k": dataIJK['k'],
            "facecolor": color,
            "hovertemplate": "Earth<extra></extra>",
        }
        # Put in main data block
        fig_dict["data"].append(data_dict)
    tocE = time.perf_counter()
    if vbose > 0:
        print(f"  -time loading Earth: {tocE - ticE:0.4f} seconds")

    # 3D===============================================================================================  DDD
    # Set layout values
    fig_dict["layout"]["height"] = 700
    fig_dict["layout"]["width"] = 800
    fig_dict["layout"]["scene_aspectmode"] = "data"
    fig_dict["layout"]["scene"] = dict(xaxis=dict(title=dict(text="X ["+scale+"]")),
                                       yaxis=dict(title=dict(text="Y ["+scale+"]")),
                                       zaxis=dict(title=dict(text="Z ["+scale+"]")))
    fig_dict["layout"]["title_text"] = txttop
    fig_dict["layout"]["showlegend"] = False
    fig_dict["layout"]["scene_camera"] = dict(center=dict(x=0, y=0, z=0))
    fig_dict["layout"]["hoverlabel_align"] = 'right'
    if ngroup > 1:
        fig_dict["layout"]["updatemenus"] = [
            {
                "buttons": [
                    {
                        "args": [None, {"frame": {"duration": 500, "redraw": True},
                                        "fromcurrent": True, "transition": {"duration": 0}}],
                        "label": "Play",
                        "method": "animate"
                    },
                    {
                        "args": [[None], {"frame": {"duration": 0, "redraw": True},
                                          "mode": "immediate",
                                          "transition": {"duration": 0}}],
                        "label": "Pause",
                        "method": "animate"
                    }
                ],
                "direction": "left",
                "pad": {"r": 10, "t": 35},
                "showactive": False,
                "type": "buttons",
                "x": 0.1,
                "xanchor": "right",
                "y": 0,
                "yanchor": "top"
            }
        ]


    # end timer                                                                               
    toc = time.perf_counter()
    if vbose > 0:
        print(f"Total time creating figure object: {toc - tic:0.4f} seconds")

    fig = go.Figure(fig_dict)

    return fig

# ===============================================================================================
# ===============================================================================================
def custom2Dpolar(datad, NS, zoom=False, vbose=1):
    """
    This function creates a custom 2D polar view satellite plot, returning a plotly figure object.
   
    Parameters
    ----------
    datad: This is a data dictionary with the data used to create the plot
    NS: string for plot type, 'N' or 'S' (any non 'S' will be considered 'N')
    zoom: logical to zoom to just > 50 degrees lat or show all
    vbose: Optional verbosity value, 0 will only print out warnings and errors. Default is 1.
    
    Returns
    -------
    fig: A plotly figure object that can then be visualized.
    
    """

    # 2DP===============================================================================================  
    # Start timer
    tic = time.perf_counter()

    # Start with error checking ...
    if 'title' not in datad:
        print("Warning, no title given for plot.")
        txttop = "No Title"
    else:
        txttop = datad['title']
    if 'var' not in datad['options']:
        print("ERROR, no variable selected to plot, returning.")
        return None
    var=datad['options']['var']
    if var == "zzztime":
        varu=""
    else:
        varu=datad[datad['sats'][0]]['vars'][var]['units']
    if 'REkm' in datad['options']:
        REkm = datad['options']['REkm']
    else:
        REkm=6.3781E3
    scale=datad['options']['position_units']
    if 'groupby' in datad['options']:
        groupby = datad['options']['groupby']
    else:
        groupby = "all"
    if 'body' in datad['options']:
        body=datad['options']['body']
    else:
        body = "none"
    if 'colorscale' in datad['options']:
        colorscale = datad['options']['colorscale']
    else:
        colorscale = "Viridis"
    if 'coord' in datad['options']:
        coord = datad['options']['coord']
    else:
        coord = ""

    # Precompute circle radius for various latitudes
    c80=np.sin(((90.-80.)/90.)*np.pi/2.)
    c70=np.sin(((90.-70.)/90.)*np.pi/2.)
    c60=np.sin(((90.-60.)/90.)*np.pi/2.)
    c50=np.sin(((90.-50.)/90.)*np.pi/2.)
    c40=np.sin(((90.-40.)/90.)*np.pi/2.)
    c30=np.sin(((90.-30.)/90.)*np.pi/2.)
    c20=np.sin(((90.-20.)/90.)*np.pi/2.)
    c10=np.sin(((90.-10.)/90.)*np.pi/2.)
    c00=1.0
    c50sq=c50**2
    # Now compute diagonal line on circle positions
    c00d=c00/np.sqrt(2.)
    c50d=c50/np.sqrt(2.)

    # set initial values used later, including loop over all sats
    xmin=0.
    xmax=0.
    ymin=0.
    ymax=0.
    zmin=0.
    zmax=0.
    cmin= 1.e99
    cmax=-1.e99
    localts=dict()
    localtimestring=dict()
    agroup=dict()
    ugroup=()
    for sat in datad['sats']:
        sPts=len(datad[sat]['vars'][datad[sat]['position_variables'][0]]['data'])
        # Set localtimestring values
        notime=False
        if 'time' in datad[sat]:
            if datad[sat]['time']['format'] == "datetime":
                localts[sat]=np.array([d.timestamp() for d in datad[sat]['time']['data']])
            elif datad[sat]['time']['format'] == "timestamp":
                localts[sat]=datad[sat]['time']['data'].copy()
            else:
                print("ERROR, Unknown time format.")
                return None
            localtimestring[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M:%S") for d in localts[sat]])
        else:
            notime=True
            if var == "zzztime":
                print("ERROR, no time given and plot var selected is time")
                return None
            localtimestring[sat]=np.array(["point "+str(i+1) for i in range(sPts)])

        # Compute mask to restrict all data in trace
        if NS == 'S':
            maskz = datad[sat]['vars']['z']['data'] <= 0.
        else:
            maskz = datad[sat]['vars']['z']['data'] >= 0.
        if zoom:
            rr = datad[sat]['vars']['x']['data']**2 + datad[sat]['vars']['y']['data']**2
            maskzoom = rr <= c50sq
            mask = maskz & maskzoom
        else:
            mask = maskz

        # Find global contour min/max
        if var == "zzztime":
            c=localts[sat]
        else:
            c=datad[sat]['vars'][var]['data']
        cmin=min(cmin,min(c[mask]))
        cmax=max(cmax,max(c[mask]))

        # Create array of possible 'groupby' value
        if groupby == "day":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d") for d in localts[sat]])
        elif groupby == "hour":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H") for d in localts[sat]])
        elif groupby == "minute":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M") for d in localts[sat]])
        elif groupby == "orbitM":
            # Satellite path crosses prime meridian
            x=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            y=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            bgroup = ['orbit'] * len(x)
            j=1
            for i in range(sPts):
                if i != 0:
                    if x[i] > 0. and (y[i]*y[i-1]) < 0.:
                        j+=1
                bgroup[i] = "orbit "+str(j)
            agroup[sat]=np.array(bgroup)
        elif groupby == "orbitE":
            # Satellite path crosses equator going North
            z=datad[sat]['vars'][datad[sat]['position_variables'][2]]['data']
            bgroup = ['orbit'] * len(z)
            j=1
            for i in range(sPts):
                if i != 0:
                    if (z[i]>0. and z[i-1]<0.):
                        j+=1
                bgroup[i] = "orbit "+str(j)
            agroup[sat]=np.array(bgroup)
        elif groupby.isdigit():
            gb=int(groupby)
            agroup[sat] = np.array(["points "+str(int(i/gb)*gb+1)+" - "+str(int(i/gb)*gb+gb) for i in range(sPts)])
        else:
            agroup[sat] = np.array(["all" for i in range(sPts)])

        # Use pandas unique function rather than numpy. Its faster and does not sort the results.
        ugroup=pd.unique(np.append(ugroup, pd.unique(agroup[sat])))

    ngroup = len(ugroup)

    # Build DUMMY data block to insert as needed.
    data_dict_dummy = {
        "type": "scatter",
        "name": "dummy", "x": [0.], "y": [0.],
        "mode": "lines", "line": {"width": 1},
        "hoverinfo": "none",
    }

    # 2DP===============================================================================================  AAA
    # make figure dictionary pieces
    fig_dict = {"data": [], "layout": {}, "frames": []}
    fig_data_saved = {"data": []}
    sliders_dict = {
        "active": 0,
        "yanchor": "top", "xanchor": "left",
        "currentvalue": {
            "prefix": "Currently showing: ",
            "visible": True,
            "xanchor": "left"
        },
        "transition": {"duration": 0},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": []
    }

    # Actual plot creation loop
    for date in ugroup:
        frame = {"data": [], "name": date}
        for sat in datad['sats']:
            # Compute mask to restrict all data in trace
            if NS == 'S':
                maskz = datad[sat]['vars']['z']['data'] <= 0.
            else:
                maskz = datad[sat]['vars']['z']['data'] >= 0.
            if zoom:
                rr = datad[sat]['vars']['x']['data']**2 + datad[sat]['vars']['y']['data']**2
                maskzoom = rr <= c50sq
                mask = maskz & maskzoom
            else:
                mask = maskz

            x=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            y=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            z=datad[sat]['vars'][datad[sat]['position_variables'][2]]['data']
            sc=1.
            if scale == "km" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "R_E":
                sc=REkm
            elif scale == "R_E" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "km":
                sc=1./REkm
            if var == "zzztime":
                c=localts[sat]
                varline=""
            else:
                c=datad[sat]['vars'][var]['data']
                varline=var+": %{marker.color:.4g} "+varu+"<br>"

            # Update position min/max values
            if date == ugroup[0]:
                xmin=min(xmin,min(x[mask]*sc))
                xmax=max(xmax,max(x[mask]*sc))
                ymin=min(ymin,min(y[mask]*sc))
                ymax=max(ymax,max(y[mask]*sc))
                zmin=min(zmin,min(z[mask]*sc))
                zmax=max(zmax,max(z[mask]*sc))

            # Compute mask to restrict all data in trace
            maskd = date == agroup[sat]
            fullmask = mask & maskd
            
            # Create hover information, including extras passed in.
            Nhv = len(datad['options']['hover_vars'])
            cd=[]
            cd.append(localtimestring[sat][fullmask])
            qline=""
            Ndv=1
            # Add Z to list
            cd.append(z[fullmask]*sc)
            qline+="Z: %{customdata["+str(Ndv)+"]:.4f} "+scale+"<br>"
            Ndv+=1
            for i in range(Nhv):
                cd.append(datad[sat]['vars'][datad['options']['hover_vars'][i]]['data'][fullmask])
                qline+=datad['options']['hover_vars'][i]+": %{customdata["+str(Ndv)+"]:.2f} "+\
                    datad[sat]['vars'][datad['options']['hover_vars'][i]]['units']+" "+\
                    datad[sat]['vars'][datad['options']['hover_vars'][i]]['coord']+"<br>"
                Ndv+=1
            cd=np.asarray(cd).T
            dateline="%{customdata[0]}<br>"

            if np.sum(fullmask) < 1:
                # Build data block with one fake point so that it looks the same
                data_dict = {
                    "type": "scatter",
                    "name": date,
                    "x": [0.], "y": [0.],
                    "mode": "markers",
                    "marker": {
                        "size": 1, "cmin": cmin, "cmax": cmax, "color": [cmin],
                        "showscale": True, "colorscale": colorscale,
                        "colorbar": { "title": "<b>"+var+"</b><br>["+varu+"]", "tickformat": ".3g" }
                    },
                    "hovertemplate": "No points in "+date+"<extra></extra>",
                }
            else:
                # Build data block with fullmask
                data_dict = {
                    "type": "scatter",
                    "name": date,
                    "x": list(x[fullmask]*sc), "y": list(y[fullmask]*sc),
                    "mode": "markers",
                    "marker": {
                        "size": 4, "cmin": cmin, "cmax": cmax, "color": list(c[fullmask]),
                        "showscale": True, "colorscale": colorscale,
                        "colorbar": { "title": "<b>"+var+"</b><br>["+varu+"]", "tickformat": ".3g" }
                    },
                    "customdata": cd,
                    "hovertemplate": "<b>"+datad[sat]['display_name']+"</b>"+
                    "<br>X: %{x:.4f} "+scale+"<br>Y: %{y:.4f} "+scale+"<br>"+
                    qline+varline+dateline+"<extra></extra>",
                }

            # If time is colorbar variable, hide labels by selecting ticks out of range
            if var == "zzztime":
                data_dict.marker.colorbar['tickvals']=(0,1)

            # Put each part of sequence in frame data block
            frame["data"].append(data_dict)

            # First in sequence, put dummy in main data block
            if date == ugroup[0]:
                fig_dict["data"].append(data_dict_dummy)
            
        fig_dict["frames"].append(frame)
        slider_step = {"args": [
            [date],
            {"frame": {"duration": 300, "redraw": True},
             "mode": "immediate",
             "transition": {"duration": 0}}
        ],
            "label": date,
            "method": "animate"}
        sliders_dict["steps"].append(slider_step)
        
    # Assemble frame and slider pieces
    fig_dict["layout"]["sliders"] = [sliders_dict]

    # 2DP===============================================================================================  BBB
    if ngroup > 1:
        for sat in datad['sats']:
            # Add trace if more than one group.
            # This shows the whole trajectory when a subsection of data is showing.
            x=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            y=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            z=datad[sat]['vars'][datad[sat]['position_variables'][2]]['data']
            sc=1.
            if scale == "km" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "R_E":
                sc=REkm
            elif scale == "R_E" and datad[sat]['vars'][datad[sat]['position_variables'][0]]['units'] == "km":
                sc=1./REkm

            # Compute mask to restrict all data in trace
            if NS == 'S':
                maskz = z <= 0.
            else:
                maskz = z >= 0.
            if zoom:
                rr = x**2 + y**2
                maskzoom = rr <= c50sq
                mask = maskz & maskzoom
            else:
                mask = maskz

            xx=np.concatenate([x,x])
            yy=np.concatenate([y,y])
            # Build new position array, element by element
            j=0
            for i in range(len(mask)):
                if mask[i]:
                    xx[j]=x[i]*sc
                    yy[j]=y[i]*sc
                    j+=1
                    if i+1 < len(mask):
                        if not mask[i+1]:
                            xx[j]=None
                            yy[j]=None
                            j+=1
            xx=np.array(xx[0:j], dtype=np.float64)
            yy=np.array(yy[0:j], dtype=np.float64)

            # Build data block
            data_dict = {
                "type": "scatter",
                "name": "positions", "x": list(xx), "y": list(yy),
                "mode": "lines", "line": {"width": 3, "color": "rgba(22,22,22,0.2)"},
                "hoverinfo": "none",
            }

            # Put into main data block
            fig_dict["data"].append(data_dict)

    # 2DP===============================================================================================  CCC
    # Build body data block
    data_dict = {
        "type": "scatter",
        "name": "positions", "x": [xmin,xmax], "y": [ymin,ymax],
        "mode": "markers", "marker": {"size": 1},
        "hoverinfo": "none",
    }
    # Put into main data block
    fig_dict["data"].append(data_dict)

    # 2DP===============================================================================================  DDD
    # Set layout values
    fig_dict["layout"]["height"] = 600
    fig_dict["layout"]["width"] = 600
    fig_dict["layout"]["scene_aspectmode"] = "data"
    fig_dict["layout"]["xaxis"] = {'visible': False, 'showticklabels': False}
    fig_dict["layout"]["yaxis"] = {'visible': False, 'showticklabels': False}
    if NS == 'S':
        if zoom:
            txttop2 = '<br>Southern Hemisphere to 50 degrees'
        else:
            txttop2 = '<br>Southern Hemisphere'
    else:
        if zoom:
            txttop2 = '<br>Northern Hemisphere to 50 degrees'
        else:
            txttop2 = '<br>Northern Hemisphere'
    fig_dict["layout"]["title_text"] = txttop+txttop2
    fig_dict["layout"]["showlegend"] = False
    fig_dict["layout"]["scene_camera"] = dict(center=dict(x=0, y=0, z=0))
    fig_dict["layout"]["hoverlabel_align"] = 'right'
    if ngroup > 1:
        fig_dict["layout"]["updatemenus"] = [
            {
                "buttons": [
                    {
                        "args": [None, {"frame": {"duration": 500, "redraw": True},
                                        "fromcurrent": True, "transition": {"duration": 0}}],
                        "label": "Play",
                        "method": "animate"
                    },
                    {
                        "args": [[None], {"frame": {"duration": 0, "redraw": True},
                                          "mode": "immediate",
                                          "transition": {"duration": 0}}],
                        "label": "Pause",
                        "method": "animate"
                    }
                ],
                "direction": "left",
                "pad": {"r": 10, "t": 75},
                "showactive": False,
                "type": "buttons",
                "x": 0.1,
                "xanchor": "right",
                "y": 0,
                "yanchor": "top"
            }
        ]

    fig = go.Figure(fig_dict)

    fig.update_xaxes(showgrid=False,scaleanchor='y')
    fig.update_yaxes(showgrid=False)
    fig.update_layout(
        paper_bgcolor='white',
        plot_bgcolor='white',
        margin=dict(l=10,b=10),
    )

    if body == "lines" and coord == "GEO":
        if zoom:
            fig.update_layout(
                shapes=[
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c80, y0=-c80, x1=c80, y1=c80),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c70, y0=-c70, x1=c70, y1=c70),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c60, y0=-c60, x1=c60, y1=c60),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50, y0=-c50, x1=c50, y1=c50),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50, y0=0., x1=c50, y1=0.),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=0., y0=-c50, x1=0., y1=c50),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50d, y0=-c50d, x1=c50d, y1=c50d),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50d, y0=c50d, x1=c50d, y1=-c50d),
                ],
            )
        else:
            fig.update_layout(
                shapes=[
                    dict(type="circle", line=dict(color="black", width=1, dash="dash"),
                         xref="x", yref="y", x0=-c80, y0=-c80, x1=c80, y1=c80),
                    dict(type="circle", line=dict(color="black", width=1, dash="dot"),
                         xref="x", yref="y", x0=-c70, y0=-c70, x1=c70, y1=c70),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c60, y0=-c60, x1=c60, y1=c60),
                    dict(type="circle", line=dict(color="black", width=1, dash="dash"),
                         xref="x", yref="y", x0=-c50, y0=-c50, x1=c50, y1=c50),
                    dict(type="circle", line=dict(color="black", width=1, dash="dot"),
                         xref="x", yref="y", x0=-c40, y0=-c40, x1=c40, y1=c40),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c30, y0=-c30, x1=c30, y1=c30),
                    dict(type="circle", line=dict(color="black", width=1, dash="dash"),
                         xref="x", yref="y", x0=-c20, y0=-c20, x1=c20, y1=c20),
                    dict(type="circle", line=dict(color="black", width=1, dash="dot"),
                         xref="x", yref="y", x0=-c10, y0=-c10, x1=c10, y1=c10),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00, y0=-c00, x1=c00, y1=c00),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00, y0=0., x1=c00, y1=0.),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=0., y0=-c00, x1=0., y1=c00),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00d, y0=-c00d, x1=c00d, y1=c00d),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00d, y0=c00d, x1=c00d, y1=-c00d),
                ],
                annotations=[
                    dict(text="60", x=-0.04, y=-(c60+0.03), showarrow=False, 
                         xref="x", yref="y", font=dict(size=8)),
                    dict(text="30", x=-0.04, y=-(c30+0.03), showarrow=False, 
                         xref="x", yref="y", font=dict(size=8)),
                ],
            )

    else:
        if zoom:
            fig.update_layout(
                shapes=[
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c80, y0=-c80, x1=c80, y1=c80),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c70, y0=-c70, x1=c70, y1=c70),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c60, y0=-c60, x1=c60, y1=c60),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50, y0=-c50, x1=c50, y1=c50),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50, y0=0., x1=c50, y1=0.),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=0., y0=-c50, x1=0., y1=c50),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50d, y0=-c50d, x1=c50d, y1=c50d),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c50d, y0=c50d, x1=c50d, y1=-c50d),
                ],
            )            
        else:
            fig.update_layout(
                shapes=[
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c60, y0=-c60, x1=c60, y1=c60),
                    dict(type="circle", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c30, y0=-c30, x1=c30, y1=c30),
                    dict(type="circle", line=dict(color="black", width=2),
                         xref="x", yref="y", x0=-c00, y0=-c00, x1=c00, y1=c00),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00, y0=0., x1=c00, y1=0.),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=0., y0=-c00, x1=0., y1=c00),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00d, y0=-c00d, x1=c00d, y1=c00d),
                    dict(type="line", line=dict(color="black", width=1),
                         xref="x", yref="y", x0=-c00d, y0=c00d, x1=c00d, y1=-c00d),
                ],
                annotations=[
                    dict(text="60", x=-0.04, y=-(c60+0.03), showarrow=False, 
                         xref="x", yref="y", font=dict(size=8)),
                    dict(text="30", x=-0.04, y=-(c30+0.03), showarrow=False, 
                         xref="x", yref="y", font=dict(size=8)),
                ],
            )

    if NS == 'S':
        fig.update_xaxes(autorange="reversed")
        
    if coord == 'GEO':
        if NS == 'S':
            if zoom:
                pos = shoreline(zlimit=-c40)
            else:
                pos = shoreline(zlimit=-0.001)
            fig.add_scatter(mode='lines', x=pos[:,0], y=pos[:,1],
                            line=dict(width=1,color='black'),
                            showlegend=False,hoverinfo='skip')
        else:
            if zoom:
                pos = shoreline(zlimit=c40)
            else:
                pos = shoreline(zlimit=0.001)
            fig.add_scatter(mode='lines', x=pos[:,0], y=pos[:,1],
                            line=dict(width=1,color='black'),
                            showlegend=False,hoverinfo='skip')

    # end timer                                                                               
    toc = time.perf_counter()
    if vbose > 0:
        print(f"Total time creating figure object: {toc - tic:0.4f} seconds")

    return fig

# ===============================================================================================
# ===============================================================================================
def custom1Dsat(datad, vbose=1):
    """
    This function creates a custom 1D satellite plot, returning a plotly figure object.
   
    Parameters
    ----------
    datad: This is a data dictionary with the data used to create the plot
    vbose: Optional verbosity value, 0 will only print out warnings and errors. Default is 1.
    
    Returns
    -------
    fig: A plotly figure object that can then be visualized.
    
    """

    if 'coord' in datad['options']:
        coord = datad['options']['coord']
    else:
        coord = ""

    # Start 1D fig
    var=datad['options']['var']
    varu=datad[datad['sats'][0]]['vars'][var]['units']
    Nhv = len(datad['options']['hover_vars'])
    
    localts=dict()
    localtimestring=dict()
    localdt=dict()
    txttop = datad['title']

    # For now this only makes the plot for the last sat in the dictionary.
    for sat in datad['sats']:
        sPts=len(datad[sat]['time']['data'])
        # Set localtimestring values
        notime=False
        if 'time' in datad[sat]:
            if datad[sat]['time']['format'] == "datetime":
                localts[sat]=np.array([d.timestamp() for d in datad[sat]['time']['data']])
            elif datad[sat]['time']['format'] == "timestamp":
                localts[sat]=datad[sat]['time']['data'].copy()
            else:
                print("ERROR, Unknown time format.")
                return None
            localtimestring[sat] = np.array([datetime.datetime.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M:%S") for d in localts[sat]])
            localdt[sat] = np.array([datetime.datetime.fromtimestamp(int(d),tz=timezone.utc) for d in localts[sat]])
        else:
            notime=True
            if var == "zzztime":
                print("ERROR, no time given and plot var selected is time")
                return None
            localtimestring[sat]=np.array(["point "+str(i+1) for i in range(sPts)])

        # Find global contour min/max
        if var == "zzztime":
            c=localts[sat]
        else:
            c=datad[sat]['vars'][var]['data']

        fig = make_subplots(rows=(Nhv+1), cols=1, shared_xaxes=True, vertical_spacing=0.04)

        fig.add_trace(go.Scatter(x=localdt[sat], y=c, name=var,
                                  mode='lines', line= dict(shape='linear', color='black'),
                                  hovertemplate=coord+'<br>'+var+': %{y:.4g}<br>%{x}<extra></extra>',
                                ),
                       row=1, col=1)
        fig.update_yaxes(title_text='<b>'+var+'</b><br>['+varu+']', exponentformat='e', row=1, col=1)
        fig.update_layout(yaxis=dict(title=dict(font=dict(size=12))))

        for i in range(Nhv):
            tmpv=datad['options']['hover_vars'][i]
            tmpu=datad[sat]['vars'][tmpv]['units']
            fig.add_trace(go.Scatter(x=localdt[sat], y=datad[sat]['vars'][tmpv]['data'], name=tmpv,
                                      mode='lines', line= dict(shape='linear', color='black'),
                                      hovertemplate=coord+'<br>'+tmpv+': %{y:.4g}<br>%{x}<extra></extra>',
                                     ),
                           row=(i+2), col=1)
#            if tmpv == "Alt":
#                tmpu=" [km]"
            if tmpv == "Lon":
#                tmpu=" [deg]"
                fig.update_yaxes(tick0=0., dtick=90., row=(i+2), col=1)
            if tmpv == "Lat":
#                tmpu=" [deg]"
                fig.update_yaxes(tick0=0., dtick=30., row=(i+2), col=1)
            ya='yaxis'+str(i+2)
            ys="dict(text='<b>"+tmpv+"</b> ["+tmpu+"]',font=dict(size=12))"
            fig['layout'][ya]['title']=eval(ys)

        fig.update_layout(height=600, width=800, title_text=txttop, showlegend = False,)

    return fig

# ===============================================================================================
# ===============================================================================================
def custom2Dsat(datad, useLT=False, vbose=1):
    '''
    This function creates a custom 2D satellite plot in lat/lon or lat/LT, returning a plotly figure object.
    
    Parameters
    ----------
    datad: This is a data dictionary with the data used to create the plot
    useLT: Optional logical to modify plot to use local time instead of longitude on the X axis.
    vbose: Optional verbosity value, 0 will only print out warnings and errors. Default is 1.
    
    Returns
    -------
    fig: A plotly figure object that can then be visualized.
    
    '''

    # 2D===============================================================================================  
    # Start timer
    tic = time.perf_counter()

    # Start with error checking ...
    if 'title' not in datad:
        print("Warning, no title given for plot.")
        txttop = "No Title"
    else:
        txttop = datad['title']
    if 'var' not in datad['options']:
        print("ERROR, no variable selected to plot, returning.")
        return None
    var=datad['options']['var']
    if var == "zzztime":
        varu=""
    else:
        varu=datad[datad['sats'][0]]['vars'][var]['units']
    if 'REkm' in datad['options']:
        REkm = datad['options']['REkm']
    else:
        REkm=6.3781E3
    scale=datad['options']['position_units']
    if 'groupby' in datad['options']:
        groupby = datad['options']['groupby']
    else:
        groupby = "all"
    if 'body' in datad['options']:
        body=datad['options']['body']
    else:
        body = "none"
    if 'colorscale' in datad['options']:
        colorscale = datad['options']['colorscale']
    else:
        colorscale = "Viridis"

    # set initial values used later, including loop over all sats
    cmin= 1.e99
    cmax=-1.e99
    localts=dict()
    localtimestring=dict()
    agroup=dict()
    ugroup=()
    for sat in datad['sats']:
        sPts=len(datad[sat]['vars'][datad[sat]['position_variables'][0]]['data'])
        # Set localtimestring values
        notime=False
        if 'time' in datad[sat]:
            if datad[sat]['time']['format'] == "datetime":
                localts[sat]=np.array([d.timestamp() for d in datad[sat]['time']['data']])
            elif datad[sat]['time']['format'] == "timestamp":
                localts[sat]=datad[sat]['time']['data'].copy()
            else:
                print("ERROR, Unknown time format.")
                return None
            localtimestring[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M:%S") for d in localts[sat]])
        else:
            notime=True
            if var == "zzztime" or useLT:
                print("ERROR, no time given and plot var selected is time")
                return None
            localtimestring[sat]=np.array(["point "+str(i+1) for i in range(sPts)])

        # Find global contour min/max
        if var == "zzztime":
            c=localts[sat]
        else:
            c=datad[sat]['vars'][var]['data']
        cmin=min(cmin,min(c))
        cmax=max(cmax,max(c))

        # Create array of possible 'groupby' value
        if groupby == "day":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d") for d in localts[sat]])
        elif groupby == "hour":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H") for d in localts[sat]])
        elif groupby == "minute":
            if notime:
                print("ERROR, no time given and groupby value is",groupby)
                return None
            agroup[sat] = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime\
                ("%Y-%m-%d %H:%M") for d in localts[sat]])
        elif groupby == "orbitM":
            # Satellite path crosses prime meridian
            lon=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            bgroup = ['orbit'] * len(lon)
            j=1
            for i in range(sPts):
                if i != 0:
                    if abs(lon[i]-lon[i-1]) > 180.:
                        j+=1
                bgroup[i] = "orbit "+str(j)
            agroup[sat]=np.array(bgroup)
        elif groupby == "orbitE":
            # Satellite path crosses equator going North
            lat=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            bgroup = ['orbit'] * len(lat)
            j=1
            for i in range(sPts):
                if i != 0:
                    if (lat[i]>0. and lat[i-1]<0.):
                        j+=1
                bgroup[i] = "orbit "+str(j)
            agroup[sat]=np.array(bgroup)
        elif groupby.isdigit():
            gb=int(groupby)
            agroup[sat] = np.array(["points "+str(int(i/gb)*gb+1)+" - "+str(int(i/gb)*gb+gb) for i in range(sPts)])
        else:
            agroup[sat] = np.array(["all" for i in range(sPts)])

        # Use pandas unique function rather than numpy. Its faster and does not sort the results.
        ugroup=pd.unique(np.append(ugroup, pd.unique(agroup[sat])))

    ngroup = len(ugroup)

    # Build DUMMY data block to insert as needed.
    data_dict_dummy = {
        "type": "scatter",
        "name": "dummy", "x": [0.], "y": [0.],
        "mode": "lines", "line": {"width": 1},
        "hoverinfo": "none",
    }

    # 2D===============================================================================================  AAA
    # make figure dictionary pieces
    fig_dict = {"data": [], "layout": {}, "frames": []}
    fig_data_saved = {"data": []}
    sliders_dict = {
        "active": 0,
        "yanchor": "top", "xanchor": "left",
        "currentvalue": {
            "prefix": "Currently showing: ",
            "visible": True,
            "xanchor": "left"
        },
        "transition": {"duration": 0},
        "pad": {"b": 10, "t": 40},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": []
    }

    # Actual plot creation loop
    for date in ugroup:
        frame = {"data": [], "name": date}
        for sat in datad['sats']:
            lon=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            lat=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']

            h = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime("%H") for d in localts[sat]])
            m = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime("%M") for d in localts[sat]])
            s = np.array([dt.fromtimestamp(int(d),tz=timezone.utc).strftime("%S.%f") for d in localts[sat]])
            lt=np.array(h, dtype=np.float64)
            for i in range(len(h)):
                lt[i]=(float(h[i]))+(float(m[i])/60.)+(float(s[i])/3600.)+lon[i]*24./360.
            lt[lt > 24.]-=24.

            if var == "zzztime":
                c=localts[sat]
                varline=""
            else:
                c=datad[sat]['vars'][var]['data']
                varline=var+": %{marker.color:.4g} "+varu+"<br>"

            # Compute mask to restrict all data in trace
            mask = date == agroup[sat]
            
            # Create hover information, including extras passed in.
            Nhv = len(datad['options']['hover_vars'])
            cd=[]
            cd.append(localtimestring[sat][mask])
            qline=""
            Ndv=1
            for i in range(Nhv):
                hovv=datad['options']['hover_vars'][i]
                if hovv != "Lat" and hovv != "Lon":
                    cd.append(datad[sat]['vars'][hovv]['data'][mask])
                    qline+=hovv+": %{customdata["+str(Ndv)+"]:.2f} "+\
                        datad[sat]['vars'][hovv]['units']+"<br>"
                    Ndv+=1
            if useLT:
                hovv='Lon'
                cd.append(datad[sat]['vars'][hovv]['data'][mask])
                qline+=hovv+": %{customdata["+str(Ndv)+"]:.2f} "+\
                    datad[sat]['vars'][hovv]['units']+"<br>"
                Ndv+=1
            else:
                cd.append(lt[mask])
                qline+="LT: %{customdata["+str(Ndv)+"]:.2f}<br>"
                Ndv+=1
            cd=np.asarray(cd).T
            dateline="%{customdata[0]}<br>"
            
            # Build data block with mask
            if useLT:
                data_dict = {
                    "type": "scatter",
                    "name": date,
                    "x": list(lt[mask]), "y": list(lat[mask]),
                    "mode": "markers",
                    "marker": {
                        "size": 9, "cmin": cmin, "cmax": cmax, "color": list(c[mask]),
                        "showscale": True, "colorscale": colorscale,
                        "colorbar": { "title": "<b>"+var+"</b><br>["+varu+"]", "tickformat": ".3g" }
                    },
                    "customdata": cd,
                    "hovertemplate": "<b>"+datad[sat]['display_name']+"</b>"+
                        "<br>LT: %{x:.4f}<br>Lat: %{y:.4f}<br>"+
                        qline+varline+dateline+"<extra></extra>",
                }
            else:
                data_dict = {
                    "type": "scatter",
                    "name": date,
                    "x": list(lon[mask]), "y": list(lat[mask]),
                    "mode": "markers",
                    "marker": {
                        "size": 9, "cmin": cmin, "cmax": cmax, "color": list(c[mask]),
                        "showscale": True, "colorscale": colorscale,
                        "colorbar": { "title": "<b>"+var+"</b><br>["+varu+"]", "tickformat": ".3g" }
                    },
                    "customdata": cd,
                    "hovertemplate": "<b>"+datad[sat]['display_name']+"</b>"+
                        "<br>Lon: %{x:.4f}<br>Lat: %{y:.4f}<br>"+
                        qline+varline+dateline+"<extra></extra>",
                }

            # If time is colorbar variable, hide labels by selecting ticks out of range
            if var == "zzztime":
                data_dict.marker.colorbar['tickvals']=(0,1)

            if ngroup == 1:
                # Put full sequence in main data block
                fig_dict["data"].append(data_dict)
                
            else:
                # Put each part of sequence in frame data block
                frame["data"].append(data_dict)
                # First in sequence, put dummy in main data block
                if date == ugroup[0]:
                    fig_dict["data"].append(data_dict_dummy)
                
        if ngroup > 1:
            fig_dict["frames"].append(frame)
        slider_step = {"args": [
            [date],
            {"frame": {"duration": 300, "redraw": True},
             "mode": "immediate",
             "transition": {"duration": 0}}
        ],
            "label": date,
            "method": "animate"}
        sliders_dict["steps"].append(slider_step)
        
    if ngroup > 1:
        # Assemble frame and slider pieces
        fig_dict["layout"]["sliders"] = [sliders_dict]

    # 2D===============================================================================================  BBB
    if ngroup > 1:
        for sat in datad['sats']:
            # Add trace if more than one group.
            # This shows the whole trajectory when a subsection of data is showing.
            lon=datad[sat]['vars'][datad[sat]['position_variables'][0]]['data']
            lat=datad[sat]['vars'][datad[sat]['position_variables'][1]]['data']
            
            # Make array larger than needed (double len(lat)), fill, and trim as needed
            xx=np.concatenate([lon,lon])
            yy=np.concatenate([lat,lat])
            # Build new position array, element by element
            j=0
            for i in range(len(lat)):
                xx[j]=lon[i]
                yy[j]=lat[i]
                j+=1
                if i < (len(lat)-1) and abs(lon[i]-lon[i+1]) > 180.:
                    L=min(abs(lon[i]),abs(360-lon[i]))
                    R=min(abs(lon[i+1]),abs(360-lon[i+1]))
                    Llat=lat[i]+(lat[i+1]-lat[i])*L/(L+R)
                    Rlat=lat[i]+(lat[i+1]-lat[i])*R/(L+R)
                    xx[j]=360.
                    yy[j]=Llat
                    xx[j+1]=None
                    yy[j+1]=None
                    xx[j+2]=0.
                    yy[j+2]=Rlat
                    j+=3
            xx=np.array(xx[0:j], dtype=np.float64)
            yy=np.array(yy[0:j], dtype=np.float64)
            
            # Build data block
            data_dict = {
                "type": "scatter",
                "name": "positions", "x": list(xx), "y": list(yy),
                "mode": "lines", "line": {"width": 1, "color": "rgba(22,22,22,0.2)"},
                "hoverinfo": "none",
            }

            # Put into main data block
            if not useLT:
                fig_dict["data"].append(data_dict)

    # 2D===============================================================================================  DDD
    # Set layout values
    if useLT:
        fig_dict["layout"]["xaxis"] = {'dtick': 3.0, 'range': [0.0, 24.0], 'tick0': 0.0,
                                       'title': {'text': '<b>LT</b> [hrs]'}}
    else:
        fig_dict["layout"]["xaxis"] = {'dtick': 30.0, 'range': [-180.0, 180.0], 'tick0': 0.0,
                                       'title': {'text': '<b>Lon</b> [deg]'}}
    fig_dict["layout"]["yaxis"] = {'dtick': 30.0, 'range': [-90.0, 90.0], 'tick0': 0.0,
                                   'title': {'text': '<b>Lat</b> [deg]'}}
    fig_dict["layout"]["title_text"] = txttop
    fig_dict["layout"]["showlegend"] = False
    fig_dict["layout"]["hoverlabel_align"] = 'right'
    if ngroup > 1:
        fig_dict["layout"]["updatemenus"] = [
            {
                "buttons": [
                    {
                        "args": [None, {"frame": {"duration": 500, "redraw": True},
                                        "fromcurrent": False, "transition": {"duration": 0}}],
                        "label": "Play",
                        "method": "animate"
                    },
                    {
                        "args": [[None], {"frame": {"duration": 0, "redraw": True},
                                          "mode": "immediate",
                                          "transition": {"duration": 0}}],
                        "label": "Pause",
                        "method": "animate"
                    }
                ],
                "direction": "left",
                "pad": {"r": 10, "t": 65},
                "showactive": False,
                "type": "buttons",
                "x": 0.1,
                "xanchor": "right",
                "y": 0,
                "yanchor": "top"
            }
        ]

    # end timer                                                                               
    toc = time.perf_counter()
    if vbose > 0:
        print(f"Total time creating figure object: {toc - tic:0.4f} seconds")

    fig = go.Figure(fig_dict)
    return fig

# %%



# -*- coding: utf-8 -*-
"""
Convert files with uniform grid to netcdf4
@author: rringuet, 2022

Date:  2020-05-05 00:00
Model: TS18
Bin:   Esw    1.1 mV/m, Bang   178 deg., tilt  13.1 deg.
Grid:  Uniform (lat_step: 1.00, lon_step: 2.00 [deg])

MLAT [deg]   MLT [hr]   Pot [kV] Vazm [deg] Vmag [m/s]
---------- ---------- ---------- ---------- ----------

"""
from glob import glob
import numpy as np
from time import perf_counter
from netCDF4 import Dataset
import re
import os

model_varnames = {"Pot": ['V', 'kV'], "Vazm": ['theta_v', 'deg'],
                  "Vmag": ['v', 'm/s'],
                  # remaining variables are time series
                  "tilt": ['theta_Btilt', "deg"], 'Esw': ['E_sw', 'mV/m'],
                  'Bang': ['theta_B', 'deg'],
                  # these are the coordinate variables
                  'MLAT': ['MLAT', 'deg'], 'MLT': ['MLT', 'hr']}


def convert_all(file_dir, pattern):
    '''Convert all files per timestep, combining data from both hemispheres if
    available.'''

    # detect grid type of output in file_dir
    ftic = perf_counter()
    test_files = sorted(glob(file_dir + pattern))
    if len(test_files) == 0:
        print('No original files found.')
        return
    gtype = grid_type(test_files[0])
    if gtype:  # 'Uniform' in line
        print('Uniform grid detected. ', end="")
        ftype = 'uni'
    else:
        print('Equal-area grid detected. ', end="")
        ftype = 'equ'
    # collect file names per timestep and convert
    files = sorted(glob(file_dir+'*'+ftype+'.txt'))
    file_patterns = np.unique([file[:-8] for file in files])
    converted_files = []
    for file_prefix in file_patterns:
        if not os.path.isfile(file_prefix+ftype+'.nc'):
            if gtype:
                converted_files.append(_toCDF(file_prefix))
            else:
                converted_files.append(_toCDFGroup(file_prefix))
    # print messages based on what happened.
    if len(converted_files) > 0:
        print(f'{len(converted_files)} files now converted into netcdf4' +
              f' files in {perf_counter()-ftic:.6f}s.')
    else:
        print('All files already converted.')
    return converted_files


def grid_type(filename):
    '''Determine grid type of data file.
    True if uniform, False if equal-area.'''

    read_obj = open(filename, 'r')
    line = read_obj.readline().strip()  # Date:  2020-05-05 00:00
    line = read_obj.readline().strip()  # Model: TS18
    # Bin:   Esw    1.1 mV/m, Bang   178 deg., tilt  13.1 deg.
    line = read_obj.readline().strip()
    # Grid:  Uniform (lat_step: 1.00, lon_step: 2.00 [deg])
    line = read_obj.readline().strip()
    read_obj.close()
    return 'Uniform' in line


# 'C:/Users/rringuet/Kamodo_WinDev1/SuperDARN/fullday/model20200505-0000.txt'
def ascii_reader(filename):
    '''Loads the data from a superdarn txt file into a nested dict. All data
    arrays are returned as 1D arrays.'''

    # open file
    read_obj = open(filename, 'r')
    # extract header
    # Date:  2020-05-05 00:00
    trash = read_obj.readline().strip()  # will get times from filename
    # Model: TS18
    model_string = read_obj.readline().strip()
    # Bin:   Esw    1.1 mV/m, Bang   178 deg., tilt  13.1 deg.
    bin_string = read_obj.readline().strip()
    # Grid:  Uniform (lat_step: 1.00, lon_step: 2.00 [deg])
    grid_string = read_obj.readline().strip()
    # empty line
    trash = read_obj.readline().strip()
    # MLAT [deg]   MLT [hr]   Pot [kV] Vazm [deg] Vmag [m/s]
    variable_keys = read_obj.readline().strip()
    # line with only dashes
    trash = read_obj.readline().strip()

    # extract info from header strings
    # '2020-05-05 00:00' = date_string[2]+' '+date_string[3]
    bin_list = bin_string[4:].split(',')
    Esw = float(bin_list[0].strip()[3:].strip().split(' ')[0])
    Bang = float(bin_list[1].strip()[4:].strip().split(' ')[0])
    tilt = float(bin_list[2].strip()[4:].strip().split(' ')[0])
    var_list = re.split(' +', variable_keys)
    header_keys = ['tilt', 'Esw', 'Bang']
    variable_keys = [item for item in var_list if '[' not in item]

    # create dictionary to store data in
    variables = {model_varnames[var][0]: {'units': model_varnames[var][-1],
                                          'data': []} for var in
                 variable_keys + header_keys}

    # store time series values
    variables[model_varnames['tilt'][0]]['data'] = tilt
    variables[model_varnames['Esw'][0]]['data'] = Esw
    variables[model_varnames['Bang'][0]]['data'] = Bang

    # store array data into dictionary
    for line in read_obj:
        vals = re.split(' +', line.strip())
        for i in range(len(variable_keys)):  # skip empty block(s) at the end
            variables[model_varnames[variable_keys[i]][0]]['data'].append(
                vals[i])
    read_obj.close()

    # convert to numpy float arrays
    for key in variables.keys():
        if isinstance(variables[key]['data'], (list)):
            variables[key]['data'] = np.array(variables[key]['data'],
                                              dtype=float)

    # add metadata
    variables['metadata'] = {'grid': grid_string[0][5:].strip(),
                             'model': model_string[0][6:].strip()}
    return variables


def df_data(df_file, verbose=False):
    '''Read in data from one hemisphere and perform latitude wrapping only.
    Logic for default grid files.'''

    # latitude grid in each file should be identical.
    # get coordinate data from file
    data = ascii_reader(df_file)
    lat = np.unique(data['MLAT']['data'])
    lon = np.unique(data['MLT']['data']) * 15.
    var1D_list = ['theta_Btilt', 'E_sw', 'theta_B']
    var3D_list = ['V', 'theta_v', 'v']

    # initialize data structures with data from file
    # some coordinate values are missing in random files. Find the matches.
    idx = [np.where(data['MLAT']['data'] == lat_val)[0] for lat_val in lat]
    idx = list(np.ravel(idx))
    variables = {var: np.reshape(data[var]['data'][idx],
                                 (lat.size, lon.size)).T
                 for var in var3D_list}  # reshape into lon/lat array
    for var in var1D_list:
        variables[var] = data[var]['data']

    # perform lat wrapping in variable data
    for var in var3D_list:
        # perform scalar averaging for pole values (latitude wrapping)
        # adding NaN on equator-side assuming the data never reaches it.
        data_shape = variables[var].shape
        total_shape = (data_shape[0], data_shape[1]+3)  # lon, lat
        tmp = np.zeros(total_shape, dtype=float)
        if min(lat) > 0:  # north pole at end of array
            tmp[:, 2:-1] = variables[var]  # copy data into grid
            tmp[:, -1] = np.mean(tmp[:, -2], axis=0)
            tmp[:, 1] = variables[var][:, 0]  # buffer row
            tmp[:, 0] = np.NaN  # add NaNs on equator side
        else:  # south pole at beginning of array
            tmp[:, 1:-2] = variables[var]  # copy data into grid
            tmp[:, 0] = np.mean(tmp[:, 1], axis=0)
            tmp[:, -2] = variables[var][:, 0]  # buffer row
            tmp[:, -1] = np.NaN  # add NaNs on equator side
        variables[var] = tmp

    # latitude wrapping in coordinate grid
    diff = min(abs(np.diff(lat)))
    if min(lat) > 0:
        # N hemisphere data
        lat = np.append(lat, 90.)
        lat = np.insert(lat, 0, [min(lat)-diff, min(lat)-diff/10.])
        # add a spot for a copy before a new NaN row
    else:
        lat = np.insert(lat, 0, -90.)
        lat = np.append(lat, [max(lat)+diff/10., max(lat)+diff])
        # add a spot for a copy before a new NaN row
    # The interpolator later assigned makes the last numerical row also NaN,
    # so adding a buffer row with a slightly different coordinate.

    # hemisphere spcific data wrangling complete. return data.
    coords = {'lon': lon, 'lat': lat}
    return variables, coords, var3D_list, var1D_list


def _toCDF(file_prefix):
    '''Reads in data from 1-2 files, writes to a netcdf4 file. Used for default
    grid data files for faster data access.'''

    # Get data from both hemispheres, one file per hemisphere
    files = sorted(glob(file_prefix+'*uni.txt'))  # N first, S second if both
    if files[0][-8] == 'N':
        variables_N, coords_N, var3D_list, var1D_list = df_data(files[0])
    if files[-1][-8] == 'S':
        variables_S, coords_S, var3D_list, var1D_list = df_data(files[-1])
    if len(files) > 2:
        print('The number of files for the northern and southern hemispheres' +
              ' are not equal. Please add the missing files and try again.')
        return False

    # Combine coordinate arrays, assuming time and lon the same
    if files[0][-8] == 'N':
        coords = coords_N
        variables = {var: variables_N[var] for var in var1D_list}
    else:
        coords = coords_S
        variables = {var: variables_S[var] for var in var1D_list}
    # Combine data and lat arrays or pick based on availability
    if len(files) > 1:
        coords['lat'] = np.append(coords_S['lat'], coords_N['lat'])
        new_shape = [len(coords['lon']), len(coords['lat'])]
        for var in var3D_list:  # south pole at beginning of array
            tmp = np.zeros(new_shape)  # north pole at end of array
            tmp[:, :len(coords_S['lat'])] = variables_S[var]
            tmp[:, len(coords_S['lat']):] = variables_N[var]
            variables[var] = tmp
    else:  # only data for one hemisphere was found, copy over data
        if files[0][-8] == 'N':
            for var in var3D_list:
                variables[var] = variables_N[var]
        else:
            for var in var3D_list:
                variables[var] = variables_S[var]

    # perform longitude wrapping in coordinate grid
    coords['lon'] = np.append(coords['lon'], 360.) - 180.
    new_shape = [len(coords['lon']), len(coords['lat'])]

    # perform lon wrapping in variable data
    for var in var3D_list:
        # repeat 180 values for -180 position
        tmp = np.zeros(new_shape, dtype=float)
        tmp[:-1, :] = variables[var]
        tmp[-1, :] = variables[var][0, :]
        variables[var] = tmp

    # Data wrangling complete. Start new output file
    cdf_filename = file_prefix+'uni.nc'
    data_out = Dataset(cdf_filename, 'w', format='NETCDF4')
    data_out.file = ''.join([f+',' for f in files]).strip(',')
    data_out.model = 'SuperDARN'
    data_out.filedate = os.path.basename(files[0])[5:].split('-')[0]
    data_out.grid = 'Uniform grid'

    # establish coordinates (lon, lat, then time open)
    # lon: create dimension and variable
    new_dim = data_out.createDimension('lon', coords['lon'].size)
    new_var = data_out.createVariable('lon', np.float32, 'lon')
    new_var[:] = coords['lon']  # store data for dimension in variable
    new_var.units = 'deg'
    # lat: create dimension and variable
    new_dim = data_out.createDimension('lat', coords['lat'].size)
    new_var = data_out.createVariable('lat', np.float32, 'lat')
    new_var[:] = coords['lat']  # store data for dimension in variable
    new_var.units = 'deg'
    # add time as a dimension only
    new_dim = data_out.createDimension('time', 1)

    # copy over variables to file
    for variable_name in variables.keys():
        # write to file
        units = [value[-1] for key, value in model_varnames.items()
                 if value[0] == variable_name][0]
        if variable_name in var3D_list:
            new_var = data_out.createVariable(variable_name, np.float32,
                                              ('lon', 'lat'))
        elif variable_name in var1D_list:
            new_var = data_out.createVariable(variable_name, np.float32,
                                              ('time'))
        else:
            continue
        new_var[:] = variables[variable_name]  # store data in variable
        new_var.units = units
    # close file
    data_out.close()
    return cdf_filename


def ea_data(ea_file, verbose=False):
    '''Read in data from one hemisphere and perform latitude wrapping only.
    Logic for equal area grid files.'''

    # get coordinate data from file
    data = ascii_reader(ea_file)
    lat = np.unique(data['MLAT']['data'])
    var1D_list = ['theta_Btilt', 'E_sw', 'theta_B']
    var3D_list = ['V', 'theta_v', 'v']

    # Longitude array is different for each latitude value. Storing in dict
    lon = {}
    for lat_val in lat:
        idx = np.where(data['MLAT']['data'] == lat_val)[0]
        lon[lat_val] = np.array(data['MLT']['data'][idx]) * 15.

    # initialize data structures with data from first file
    var1D_list = ['theta_Btilt', 'E_sw', 'theta_B']
    var3D_list = ['V', 'theta_v', 'v']
    variables = {var: data[var]['data'] for var in var1D_list}
    for var in var3D_list:
        variables[var] = {}
        for lat_val in lat:
            idx = np.where(data['MLAT']['data'] == lat_val)[0]
            variables[var][lat_val] = np.array(data[var]['data'][idx]).T
    coords = {'lon': lon, 'lat': lat}

    # The interpolator later assigned makes the last numerical row also NaN,
    # so adding a buffer row with a slightly different latitude value.
    # Also adding a spot for the averaging at the pole.
    diff = min(abs(np.diff(lat)))
    if min(lat) > 0:
        # N hemisphere data
        lat_equator, lat_pole = lat[0], lat[-1]
        lat = np.append(lat, 90.)
        new_vals = [min(lat)-diff, min(lat)-diff/10.]
        lat = np.insert(lat, 0, new_vals)
        # added a spot for a copy before a new NaN row
        new_vals.append(90)
    else:
        lat_equator, lat_pole = lat[-1], lat[0]
        lat = np.insert(lat, 0, -90.)
        new_vals = [max(lat)+diff/10., max(lat)+diff]
        lat = np.append(lat, new_vals)
        new_vals.reverse()  # want diff/10 value second in list
        # added a spot for a copy before a new NaN row
        new_vals.append(-90)
    coords['lat'] = lat  # update lat values in dictionary

    # adding coordinates to lon dictionary for new latitude values
    lon[new_vals[0]] = lon[lat_equator]  # same grid as equator-most lat value
    lon[new_vals[1]] = lon[lat_equator]  # for two buffer rows
    lon[new_vals[2]] = np.array([-180., 0., 180.])  # range for interpolations
    coords['lon'] = lon  # update lon values in dictionary

    # perform lat wrapping in variable data
    for var in var3D_list:
        # perform scalar averaging for pole values (latitude wrapping)
        variables[var][new_vals[2]] = np.mean(variables[var][lat_pole])

        # addind NaN on equator-side assuming the data never reaches it.
        variables[var][new_vals[1]] = variables[var][lat_equator]  # buffer row
        variables[var][new_vals[0]] = np.NaN * \
            np.ones(shape=variables[var][lat_equator].shape)

    # hemisphere spcific data wrangling complete. return data.
    return variables, coords, var3D_list, var1D_list


def _toCDFGroup(file_prefix):
    '''Reads in data from 1-2 files, writes to a netcdf4 file. Used for
    equal-area grid data files for faster data access.'''

    # Get data from both hemispheres, one file per hemisphere
    files = sorted(glob(file_prefix+'*equ.txt'))  # N first, S second if both
    if files[0][-8] == 'N':
        variables_N, coords_N, var3D_list, var1D_list = ea_data(files[0])
    if files[-1][-8] == 'S':
        variables_S, coords_S, var3D_list, var1D_list = ea_data(files[-1])
    if len(files) > 2:
        print('The number of files for the northern and southern hemispheres' +
              ' are not equal. Please add the missing files and try again.')
        return False

    # Combine coordinate arrays, assuming time and lon the same
    if files[0][-8] == 'N':
        coords = coords_N
        variables = {var: variables_N[var] for var in var1D_list}
    else:
        coords = coords_S
        variables = {var: variables_S[var] for var in var1D_list}
    # Combine data and lat arrays or pick based on availability
    if len(files) > 1:
        coords['lat'] = np.append(coords_S['lat'], coords_N['lat'])
        coords['lon'] = coords_N['lon']  # nested dictionary with keys=lat_vals
        for lat_val in coords_S['lat']:
            coords['lon'][lat_val] = coords_S['lon'][lat_val]
        for var in var3D_list:  # south pole at beginning of array
            variables[var] = variables_N[var]
            for lat_val in coords_S['lat']:
                variables[var][lat_val] = variables_S[var][lat_val]
    else:  # only data for one hemisphere was found, copy over data
        if files[0][-8] == 'N':
            for var in var3D_list:
                variables[var] = variables_N[var]
        else:
            for var in var3D_list:
                variables[var] = variables_S[var]

    # perform longitude wrapping in coordinate grid and variable data
    for lat_val in coords['lat'][1:-1]:  # skip poles because already correct
        # Deal with lat_val specific coordinate grid first
        lon_flag = ''
        # new lon grids are linked and will change together
        if max(coords['lon'][lat_val]) > 180.:
            coords['lon'][lat_val] -= 180.
        if min(coords['lon'][lat_val]) > -180.:  # extend left
            diff = min(abs(np.diff(coords['lon'][lat_val])))
            coords['lon'][lat_val] = np.insert(coords['lon'][lat_val], 0,
                                               min(coords['lon'][lat_val]) -
                                               diff)
            lon_flag += 'L'
        if max(coords['lon'][lat_val]) < 180.:  # extend right
            diff = min(abs(np.diff(coords['lon'][lat_val])))
            coords['lon'][lat_val] = np.append(coords['lon'][lat_val],
                                               max(coords['lon'][lat_val]) +
                                               diff)
            lon_flag += 'R'
        new_shape = (len(coords['lon'][lat_val]))

        # perform lon wrapping in variable data
        if lon_flag == '':
            continue
        if lon_flag == 'L':
            for var in var3D_list:
                tmp = np.zeros(new_shape, dtype=float)
                tmp[1:] = variables[var][lat_val]
                tmp[0] = variables[var][lat_val][-1]
                variables[var][lat_val] = tmp
        elif lon_flag == 'R':
            for var in var3D_list:
                tmp = np.zeros(new_shape, dtype=float)
                tmp[:-1] = variables[var][lat_val]
                tmp[-1] = variables[var][lat_val][0]
                variables[var][lat_val] = tmp
        elif lon_flag == 'LR':
            for var in var3D_list:
                tmp = np.zeros(new_shape, dtype=float)
                tmp[1:-1] = variables[var][lat_val]
                tmp[0] = variables[var][lat_val][-1]
                tmp[-1] = variables[var][lat_val][0]
                variables[var][lat_val] = tmp

    # Data wrangling complete. Start new output file
    cdf_filename = file_prefix+'equ.nc'
    data_out = Dataset(cdf_filename, 'w', format='NETCDF4')
    data_out.file = ''.join([f+',' for f in files]).strip(',')
    data_out.model = 'SuperDARN'
    data_out.filedate = os.path.basename(files[0])[5:].split('-')[0]
    data_out.grid = 'Equal area grid'

    # establish coordinates (lat, then time open)
    # lat: create dimension and variable
    new_dim = data_out.createDimension('lat', coords['lat'].size)
    new_var = data_out.createVariable('lat', np.float32, 'lat')
    new_var[:] = coords['lat']  # store data for dimension in variable
    new_var.units = 'deg'
    # time: create dimension only
    new_dim = data_out.createDimension('time', 1)

    # copy over 1D variables to file
    for variable_name in var1D_list:
        new_var = data_out.createVariable(variable_name, np.float32,
                                          ('time'))
        new_var[:] = variables[variable_name]  # store data in variable
        units = [value[-1] for key, value in model_varnames.items()
                 if value[0] == variable_name][0]
        new_var.units = units

    # Group lon and 3D data by lat values
    for lat_val in coords['lat']:
        if lat_val < 0:
            lat_name = str(lat_val).replace('.', '_').replace('-', 'n')
        else:
            lat_name = 'p'+str(lat_val).replace('.', '_')
        new_group = data_out.createGroup(lat_name)
        # lon: create dimension and variable
        new_dim = new_group.createDimension('lon', coords['lon'][lat_val].size)
        new_var = new_group.createVariable('lon', np.float32, 'lon')
        new_var[:] = coords['lon'][lat_val]  # store data for dimension
        new_var.units = 'deg'

        # copy over 3D variables to file
        for variable_name in var3D_list:
            # output to file
            new_var = new_group.createVariable(variable_name, np.float32,
                                               ('lon'))
            new_data = variables[variable_name][lat_val]
            new_var[:] = new_data  # store data in variable
            units = [value[-1] for key, value in model_varnames.items()
                     if value[0] == variable_name][0]
            new_var.units = units

    # close file
    data_out.close()
    return cdf_filename



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 20 11:57:13 2022

@author: rringuet
"""
from csv import reader
from datetime import datetime, timezone
from netCDF4 import Dataset
import numpy as np
from glob import glob
from os.path import basename, isfile


def convert_all(file_dir):
    '''Converting all files to netCDF4.'''

    files = sorted(glob(file_dir + '*.txt'))  # want YYYYMMDD
    dates = np.unique([basename(f).split('.')[0][-8:] for f in files])
    # print(dates)
    file_patterns = np.unique([basename(f).split('.')[0][:-9] for f in files])
    # file(s) could be from either hemisphere, but only one file naming pattern
    new_pattern = np.unique([''.join(*[basename(p).split('_N')]) if '_N' in p
                             else ''.join(*[basename(p).split('_S')])
                             for p in file_patterns])[0]  # only one pattern
    for d in dates:
        new_file = file_dir + new_pattern + '_' + d + '.nc'
        if not isfile(new_file):  # only convert if converted file not found
            to_CDF(new_file[:-3])  # slice off '.nc'
    return True


def adelphi_data_hemisphere(file_name):
    '''Read in data from given file and return in a dictionary of arrays.'''

    # read in data from file
    file_obj = open(file_name, 'r')
    csv_reader = reader(file_obj, delimiter='\n')  # splits file by \n char

    # pull in date and time
    line = next(csv_reader)
    data = {'Time': [float(line[0][9:].strip().split(' ')[0])]}  # in hours
    year, month, day = int(line[0][:4]), int(line[0][4:6]), int(line[0][6:8])
    dt_str = datetime(year, month, day, 0, 0, 0, tzinfo=timezone.utc
                      ).strftime('%Y-%m-%d %H:%M:%S')

    # set up data structure with next two lines
    line = next(csv_reader)
    data['MLT'] = [float(line[0].split('=')[1].split('Hour')[0].strip())]
    line = next(csv_reader)
    keys = [key for key in line[0].split(' ') if key != '']  # MLAT, PED, etc
    for key in keys:
        data[key] = [[[]]]

    # read in data
    time_idx, lon_idx = 0, 0
    for line in csv_reader:
        if line[0][:4] == dt_str[:4]:  # retrieve time value
            data['Time'].append(float(line[0][9:].strip().split(' ')[0]))
            time_idx += 1
            lon_idx = -1
            for i in range(len(keys)):
                data[keys[i]].append([])
        elif line[0][:3] == 'MLT':  # retrive coordinate value
            data['MLT'].append(float(line[0].split('=')[1].split('Hour')[0].
                                     strip()))
            lon_idx += 1
            for i in range(len(keys)):
                data[keys[i]][time_idx].append([])
        elif line[0][:4] == 'MLAT':  # skip repeat of keys
            continue
        else:  # read in data values
            # check for numbers running against each other, insert space if yes
            items = line[0][1:].split('-')  # avoid preceding - for S hemispher
            if len(items) > 1:
                # print(items)
                for i in range(len(items[1:])):
                    if items[i][-1] != ' ':
                        items[i] += ' '
                line[0] = line[0][0] + ''.join(['-'+item for item in items
                                                ])[1:]

            # put data in correct list
            values = [float(value) for value in line[0].split(' ') if
                      value != '']
            for i in range(len(keys)):
                data[keys[i]][time_idx][lon_idx].append(values[i])
    file_obj.close()

    # convert to arrays an reshape variable data
    data['Time'] = np.array(data['Time'])  # time in hours
    data['MLT'] = np.unique(np.array(data['MLT'])) * 15.  # magnetic longitude
    data['MLAT'] = np.array(data['MLAT'][0][0])  # magnetic latitude
    # in the data MLAT is fastest changing, then MLT, then Time
    for key in keys[1:]:
        data[key] = np.array(data[key])
    return data, dt_str


def combine_hemispheres(north_file, south_file):
    '''Given the filenames for the northern and southern hemispheres, combine
    the data into a single dictionary and return.'''

    data_N, date_str = adelphi_data_hemisphere(north_file)
    data_S, date_str = adelphi_data_hemisphere(south_file)
    key_list = [key for key in data_N.keys() if key not in ['Time', 'MLT',
                                                            'MLAT']]
    # Combine coordinate grids, carefully piecing together the latitudes
    # not extending towards the poles because the model fills in zeros there
    len_lat = len(data_N['MLAT'])
    data = {'Time': data_N['Time'], 'Lon': np.append(data_N['MLT'], 360.)-180.}
    data['Lat'] = np.zeros(len_lat*2+4)  # need NaN and buffer rows at equator
    data['Lat'][:len_lat] = np.flip(data_S['MLAT'])
    # add buffer rows for S hemisphere
    data['Lat'][len_lat] = data['Lat'][len_lat-1] +\
        abs(np.diff(data_S['MLAT'])[0])/10.
    data['Lat'][len_lat+1] = data['Lat'][len_lat-1] +\
        abs(np.diff(data_S['MLAT'])[0])
    data['Lat'][len_lat+4:] = data_N['MLAT']  # add in N hemi lats
    data['Lat'][len_lat+3] = data['Lat'][len_lat+4] -\
        np.diff(data_N['MLAT'])[0]/10.  # add buffer rows for N hemisphere
    data['Lat'][len_lat+2] = data['Lat'][len_lat+4] -\
        np.diff(data_N['MLAT'])[0]

    # combine data from different hemispheres
    new_shape = (len(data['Time']), len(data['Lon']), len(data['Lat']))
    for key in key_list:
        # pull in data and set buffer rows on equator side for both hemispheres
        data[key] = np.zeros(new_shape)
        data[key][:, 1:, :len_lat] = np.flip(data_S[key], axis=2)  # S data
        data[key][:, 1:, len_lat] = data[key][:, 1:, len_lat-1]  # value buffer
        data[key][:, 1:, len_lat+1:len_lat+3] = \
            np.tile(np.NaN, (new_shape[0], new_shape[1]-1, 2))  # NaN buffer
        data[key][:, 1:, len_lat+3] = data[key][:, 1:, len_lat+4]  # value buff
        data[key][:, 1:, len_lat+4:] = data_N[key]  # N hemisphere data

        # wrap in longitude
        data[key][:, 0, :] = data[key][:, -1, :]

        # The # of rows cut off is different for each variable, so avoiding
        # slicing off zeros at each pole, instead replace with NaNs.
        # find first latitude row that is nonzero, only near poles
        # this purposefully ignores the buffer rows near the equator
        # south pole at beginning
        SP_idx, slice_idx = 0, []
        zero_check = np.count_nonzero(data[key][:, :, SP_idx])
        while zero_check == 0:
            slice_idx.append(SP_idx)
            SP_idx += 1
            zero_check = np.count_nonzero(data[key][:, :, SP_idx])
        # north pole at the end
        NP_idx = -1
        zero_check = np.count_nonzero(data[key][:, :, NP_idx])
        while zero_check == 0:
            slice_idx.append(NP_idx)
            NP_idx -= 1
            zero_check = np.count_nonzero(data[key][:, :, NP_idx])
        # replace 'extra' latitude rows from data with NaNs
        data[key][:, :, slice_idx] = np.tile(np.NaN, (
            new_shape[0], new_shape[1], len(slice_idx)))

    # data wrangling complete, return data dictionary
    return data, date_str


def to_CDF(file_prefix):
    '''Given the file_prefix of form ADELPHI_2D_MAG_YYYMMDD, find and combine
    data from both hemispheres into a netCDF4 file.'''
    from time import perf_counter

    print('Converted data file not found. Converting files with ' +
          f'{file_prefix} naming pattern to a netCDF4 file.')  # , end="")
    north_file = file_prefix[:-13]+'_N'+file_prefix[-13:]+'.txt'
    south_file = file_prefix[:-13]+'_S'+file_prefix[-13:]+'.txt'
    time0 = perf_counter()
    data, date_str = combine_hemispheres(north_file, south_file)  # ~ 18-19 s
    time1 = perf_counter()
    print('Time to read and combine the data sets:', time1-time0)

    # file_prefix = file_dir + 'ADELPHI_2D_MAG_20130317'
    cdf_filename = file_prefix + '.nc'
    data_out = Dataset(cdf_filename, 'w', format='NETCDF4')
    data_out.file = north_file + ',' + south_file
    data_out.model = 'ADELPHI'
    data_out.filedate = date_str

    # establish coordinates (time, lon, lat)
    # time: create dimension and variable
    new_dim = data_out.createDimension('time', data['Time'].size)
    new_var = data_out.createVariable('time', np.float32, 'time')
    new_var[:] = data['Time']
    new_var.units = 'hr'
    # lon: create dimension and variable
    new_dim = data_out.createDimension('lon', data['Lon'].size)
    new_var = data_out.createVariable('lon', np.float32, 'lon')
    new_var[:] = data['Lon']  # store data for dimension in variable
    new_var.units = 'deg'
    # lat: create dimension and variable
    new_dim = data_out.createDimension('lat', data['Lat'].size)
    new_var = data_out.createVariable('lat', np.float32, 'lat')
    new_var[:] = data['Lat']  # store data for dimension in variable
    new_var.units = 'deg'

    # copy over variables to file
    var_list = [key for key in data.keys() if key not in
                ['Time', 'Lon', 'Lat']]
    for variable_name in var_list:
        new_var = data_out.createVariable(variable_name, np.float32,
                                          ('time', 'lon', 'lat'))
        new_var[:] = data[variable_name]  # store data in variable
        new_var.units = 'V'

    # close file
    data_out.close()
    print('Time to create netCDF4 file:', perf_counter()-time0)
    return cdf_filename



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 23 17:09:14 2021
@author: rringuet
Most needed functions to support the SatelliteFlythrough software.
"""
from numpy import vectorize, array, diff, where, unique
from time import perf_counter
from datetime import datetime, timedelta
from os.path import isfile
import kamodo_ccmc.flythrough.model_wrapper as MW
from kamodo_ccmc.readers.reader_utilities import read_timelist
from kamodo_ccmc.flythrough.utils import ConvertCoord


@vectorize
def ts_to_hrs(utcts, filedate):
    '''Convert array of timestamps to hours since midnight of filedate.'''
    return (utcts - filedate.timestamp())/3600.


def ts_to_ISOstring(utc_ts):
    '''Convert timestamp to string of format 2017-05-28T00:00:00'''
    return datetime.utcfromtimestamp(utc_ts).isoformat()


def sat_tracks(sat_time, c1, c2, c3, z_dependencies, verbose=False):
    '''Calculate satellite tracks for interpolation'''
    # Create satellite tracks with appropriate inputs
    sat_track = {}  # initialize list of satellite tracks
    if '3D' in z_dependencies.keys():
        sat_track['3D'] = array([sat_time, c1, c2]).T
    if '4D' in z_dependencies.keys():
        sat_track['4D'] = array([sat_time, c1, c2, c3]).T
    return sat_track


def Model_FlyAway(reader, file_dir, variable_list, sat_time, c1, c2, c3,
                  z_dependencies, verbose=False):
    '''Perform flythrough for one day of data and one coordinate system.'''

    # create kamodo object, initialize some variables
    var_list = variable_list.copy()  # save copy before it gets altered
    kamodo_object = reader(file_dir, variables_requested=variable_list,
                           gridded_int=False)

    # remove requested variables not found in data from var_list and z_dependen
    newvar_list = [var for var in var_list if var in
                   kamodo_object.variables.keys()]
    z_dependencies = {key: [var for var in value if var in newvar_list]
                      for key, value in z_dependencies.items()}
    # e.g. {'3D': a list of variables, '4D': a list of variables}
    # used as a mapping of coordinate types to variables

    # create satellite tracks of types needed based on vertical dependencies
    sat_track = sat_tracks(sat_time, c1, c2, c3, z_dependencies,
                           verbose=verbose)

    # retrieve interpolator and interpolate data for each variable, using track
    #   type appropriate for each variable.
    results = {var: kamodo_object[var](sat_track[[key for key, value in
                                                  z_dependencies.items() if var
                                                  in value][0]]) for var in
               newvar_list}
    del kamodo_object   # save memory
    return results


def coordinate_systems(model, sat_time, c1, c2, c3, variable_list, coord_type,
                       coord_grid):
    '''Detect what coordinate system is needed per variable, convert and return
    per type.
    Returns a dictionary. Keys are the coordinate systems. Values are
    0: a list of the variable names needing those coordinates
    1: x/lon 1D array, converted to the coordinate system key
    2: y/lat 1D array, converted to the coordinate system key
    3: z/rad/height 1D array, converted to the coordinate system key
    4: z_dependencies dictionary {3D: ([t, c1, c2]).T, 4D: ([t, c1, c2, c3]).T}
    '''

    # determine coordinate types needed
    # convert to alternative coordinates if necessary
    var_dict = MW.Model_Variables(model, return_dict=True)
    # {varname: [desc, int, coord_name, grid_type, coord_list, unit]}
    var_coord_strs = unique([value[2] + ',' + value[3] for
                             key, value in var_dict.items() if key in
                             variable_list])
    # 'SPH,sph','MAG,sph','GDZ,sph', etc
    if len(var_coord_strs) != 1 or var_coord_strs[0] != (coord_type + ',' +
                                                         coord_grid):
        # then coordinate conversion needed
        new_coords = {coord_name: [[key for key, value in var_dict.items()
                                    if (value[2].split('_')[0]+','+value[3] ==
                                        coord_name) and key in variable_list]]
                      for coord_name in var_coord_strs}  # key is 'name,type'
        # first value is a list of the variable names needing those coordinates
        for key in new_coords.keys():   # e.g. key= 'GDZ,sph'
            # convert to needed coordinates
            alt_c1, alt_c2, alt_c3, units_out = \
                    ConvertCoord(sat_time, c1, c2, c3, coord_type, coord_grid,
                                 *key.split(','))
            new_coords[key].extend([alt_c1, alt_c2, alt_c3])  # elements 1,2,3
        # second, third and fourth values are 1D position arrays
    else:
        new_coords = {coord_type+','+coord_grid: [variable_list, c1, c2, c3]}

    # determine z_dependency of relevant variables for each coordinate system
    for key in new_coords.keys():
        z_dependencies = {}
        coord_lengths = [len(value[4]) for keyv, value in var_dict.items() if
                         keyv in new_coords[key][0]]
        if 3 in coord_lengths:  # determine which variables are 3D
            z_dependencies['3D'] = [keyv for keyv, value in var_dict.items() if
                                    len(value[4]) == 3 and keyv in
                                    new_coords[key][0]]
        if 4 in coord_lengths:  # determine which variables are 4D
            z_dependencies['4D'] = [keyv for keyv, value in var_dict.items() if
                                    len(value[4]) == 4 and keyv in
                                    new_coords[key][0]]
        new_coords[key].append(z_dependencies)  # element 5
        # e.g. {'3D': a list of variables, '4D': a list of variables}
        # used as a mapping of coordinate types to variables
    return new_coords


def Model_SatelliteFlythrough(model, file_dir, variable_list, sat_time, c1, c2,
                              c3, coord_type, coord_grid, verbose=False):
    '''
    Execute flythrough for model data. Returns results_dict.
    results_dict is a dictionary of the interpolated data for the entire data
        set sorted by variable name.
    file_dir is a string indicating where the data files are located.
    variable_list is a list of strings of the desired variable names.
    sat_time is an array of timestamp values.
    c1, c2, c3 = x, y, z or lon, lat, height (or radius)
    if x, y, z, then must be in R_E units
    if radius -> R_E. if height -> km
    '''

    # Check that sat data is all the same length, will error if not
    if max(diff(array([len(sat_time), len(c3), len(c2), len(c1)]))) > 0:
        raise AttributeError('Satellite arrays or lists must all be the ' +
                             'same length. Current array lengths are ' +
                             f'{len(sat_time)}, {len(c1)}, {len(c2)}, and ' +
                             f'{len(c3)}')

    # initialize model reader and times dictionary
    reader = MW.Model_Reader(model)  # Kamodo gets imported here
    start_utcts, end_utcts, filedate = File_UTCTimes(model, file_dir)
    
    # cut off trajectory times not found in data in file_dir
    idx = where((sat_time >= start_utcts) & (sat_time <= end_utcts))[0]
    results_dict = {'utc_time': sat_time[idx], 'c1': c1[idx],
                    'c2': c2[idx], 'c3': c3[idx], 'net_idx': idx}
    # net_idx for comparison with other data from real satellite
    if len(idx) < 1:
        print('None of the given times are in the model output files.')
        return results_dict
    if len(idx) < len(sat_time):
        print(f'{len(sat_time)-len(idx)} times are not in model output files' +
              ' and are excluded from the flythrough.')
    
    # perform coordinate conversions and sort variables by coordinate systems
    coord_dict = coordinate_systems(model, sat_time[idx], c1[idx], c2[idx],
                                    c3[idx], variable_list, coord_type,
                                    coord_grid)

    # perform flythroughs
    if verbose:
        print('Interpolating through model data...', end="")
    interp_time = perf_counter()
    for key in coord_dict.keys():  # keys = coordinate system names
        # interpolate requested data for each coordinate system
        # reader, file_dir, variable_list, sat_time in hrs, c1, c2, c3,
        #   z_dependencies
        
        results = Model_FlyAway(reader, file_dir, coord_dict[key][0],
                                ts_to_hrs(sat_time[idx], filedate),
                                coord_dict[key][1], coord_dict[key][2],
                                coord_dict[key][3], coord_dict[key][4],
                                verbose=verbose)

        # collect interpolated data into the same dictionary
        for var in results.keys():
            results_dict[var] = results[var]
    if verbose:
        print(f'done in {perf_counter()-interp_time:.5f} s.')
    return results_dict


def File_UTCTimes(model, file_dir):
    '''Returns the min start utc timestamp, the max end utc timestamp, and the
    datetime object containing midnight of the start date of the data. Creates
    preprocessed files if needed.'''

    # get times dictionary and datetime filedate object from files
    list_file = file_dir + model +'_list.txt'
    time_file = file_dir + model +'_times.txt'
    if isfile(list_file):
        times, tmp, filedate, tmp = read_timelist(time_file, list_file)
    else:
        reader = MW.Model_Reader(model)
        ko = reader(file_dir, filetime=True)  # creates any preprocessed files
        times, filedate = ko.times, ko.filedate
        del ko
        
    # calculate minimum start time
    start, end = [], []
    for p in times.keys():
        start.append(times[p]['start'][0])
        end.append(times[p]['end'][-1])
    start_utcts = (filedate + timedelta(hours=float(min(start)))).timestamp()
    end_utcts = (filedate + timedelta(hours=float(max(end)))).timestamp()
    return start_utcts, end_utcts, filedate
        
